//=============================================================================
// RPG Maker MZ - Ironman Mode
// ----------------------------------------------------------------------------
// (C)2025 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 2025/03/18
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Provides Ironman Mode functionality
 * @author aura-dev
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help ironman.js
 *
 * - Disables autosaves and locks the player to a chosen save slot upon ironman activation.
 * - Saving the game will send the player to the Title Screen.
 * - Sets a flag upon saving the game and unsets it when loading. Said flag can be used in
 * order to handle savescumming.
 * 
 * @param ironmanSaveLabel_cte
 * @type text
 * @text Ironman Save Label
 * @desc The text that will be shown for the chosen save slot.
 * @default [IRONMAN]
 * 
 * @param onReloadCommonEventID
 * @type number
 * @text onReload CommonEvent ID
 * @desc If the player quits the game without proper saving and then loads the save, this common event will be called.
 * 
 * @command activate
 * @text Activate
 * @desc Activates ironman mode for the given save slot
 * 
 * @arg saveID
 * @type text
 * @text Save ID
 * @desc The ID of the save slot. Argument will be evaluated.
 */

window.AuraMZ = window.AuraMZ || {};
AuraMZ.Ironman = {};

(($) => {
	PLUGIN_ID = "ironman";
	$.Params = PluginManager.parameters(PLUGIN_ID);
	$.onReloadCommonEventID = parseInt($.Params["onReloadCommonEventID"]);

	// Setup localizable strings
	Object.defineProperties($, {
		"ironmanSaveLabel_cte": { get: () => TextManager.message(`auramz/${PLUGIN_ID}-ironmanSaveLabel_cte`) || $.Params["ironmanSaveLabel_cte"] }
	});

	// Command to actviate ironman mode
	PluginManager.registerCommand(PLUGIN_ID, "activate", args => {
		const saveID = eval(args.saveID);
		$gameSystem.enableIronman(saveID);
		$gameSystem._ironmanSaveFlag = true;
		DataManager.saveGame(saveID);
	});

	// Checks if ironman is enabled
	Game_System.prototype.isIronmanEnabled = function() {
		return this._ironmanSaveID > 0;
	};

	// Enable ironman using the given save ID
	Game_System.prototype.enableIronman = function(ironmanSaveID) {
		this._ironmanSaveID = ironmanSaveID;
	};

	// Inject logic to save if a save is in ironman on the info level
	$.DataManager_makeSavefileInfo = DataManager.makeSavefileInfo;
	DataManager.makeSavefileInfo = function() {
		const info = $.DataManager_makeSavefileInfo.call(this);
		info.ironman = $gameSystem.isIronmanEnabled();
		return info;
	};

	// Inject logic to mark iron man saves through a name suffix
	$.DataManager_savefileInfo = DataManager.savefileInfo;
	DataManager.savefileInfo = function(savefileId) {
		const info = $.DataManager_savefileInfo.call(this, savefileId);
		if (info != null && info.ironman) {
			info.saveName = `${TextManager.file} #` + savefileId + " " + $.ironmanSaveLabel_cte;
		}
		return info;
	};

	// Inject logic to disable autosaving on ironman
	$.Game_System_isAutosaveEnabled = Game_System.prototype.isAutosaveEnabled;
	Game_System.prototype.isAutosaveEnabled = function() {
		return $.Game_System_isAutosaveEnabled.call(this) && !this.isIronmanEnabled();
	};

	// Inject logic to disable saving during dialogue
	$.Game_System_isSaveEnabled = Game_System.prototype.isSaveEnabled;
	Game_System.prototype.isSaveEnabled = function() {
		if (this.isIronmanEnabled() && $gameMap.isEventRunning()) {
			return false;
		}
		return $.Game_System_isSaveEnabled.call(this);
	};

	// Inject logic to restrict saving to the dedicated ironman save file
	$.Scene_File_isSavefileEnabled = Scene_File.prototype.isSavefileEnabled;
	Scene_File.prototype.isSavefileEnabled = function(savefileId) {
		if (this.mode() == "save" && $gameSystem.isIronmanEnabled()) {
			return $gameSystem._ironmanSaveID == savefileId;
		}

		return $.Scene_File_isSavefileEnabled.call(this, savefileId);
	};

	// Inject logic to mark the ironman save as properly saved
	$.Scene_Save_executeSave = Scene_Save.prototype.executeSave;
	Scene_Save.prototype.executeSave = function(savefileId) {
		if ($gameSystem.isIronmanEnabled()) {
			$gameSystem._ironmanSaveFlag = true;
		}

		$.Scene_Save_executeSave.call(this, savefileId);
	};

	// Inject logic to go to the title screen upon saving in ironman
	$.Scene_Save_onSaveSuccess = Scene_Save.prototype.onSaveSuccess;
	Scene_Save.prototype.onSaveSuccess = function(savefileId) {
		$.Scene_Save_onSaveSuccess.call(this, savefileId);
		if ($gameSystem.isIronmanEnabled()) {
			SceneManager.goto(Scene_Title);
		}
	};

	// Inject logic to check if the player exited by saving
	$.Game_System_onAfterLoad = Game_System.prototype.onAfterLoad;
	Game_System.prototype.onAfterLoad = function() {
		if (this.isIronmanEnabled()) {
			this._checkIronmanFlag = true;
		}

		$.Game_System_onAfterLoad.call(this);
	};

	// Inject logic to execute the reload common event if the
	// player has been detected to perform a reload on the save
	$.Scene_Map_onMapLoaded = Scene_Map.prototype.onMapLoaded;
	Scene_Map.prototype.onMapLoaded = function() {
		$.Scene_Map_onMapLoaded.call(this);

		if ($gameSystem._checkIronmanFlag) {
			if (!$gameSystem._ironmanSaveFlag && $.onReloadCommonEventID > 0) {
				$gameTemp.reserveCommonEvent($.onReloadCommonEventID);
			}
			$gameSystem._ironmanSaveFlag = false;
			$gameSystem._checkIronmanFlag = false;
			DataManager.saveGame($gameSystem._ironmanSaveID);
		}
	};
	
	// Inject check to not open the save menu when if in ironman mode
	$.Game_Interpreter_command352 = Game_Interpreter.prototype.command352;
	Game_Interpreter.prototype.command352 = function() {
		if ($gameSystem.isIronmanEnabled()) return true;
		return $.Game_Interpreter_command352.call(this);
	};
})(AuraMZ.Ironman);

