//=============================================================================
// RPG Maker MZ - Mod Loader
// ----------------------------------------------------------------------------
// (C)2025 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 2025/02/18
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Mod Loader
 * @author aura-dev
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help mod_loader.js
 *
 * Redirects requests for resources to resources from the modes/ folder.
 * For each mod, the mod loader expects a file called "Meta.json":
 * {
 * 	"version": "VERSION",
 * 	"enabled": true iff the mod should be loaded
 * 	"decryptedImages": true iff the mod uses plain images
 * 	"decryptedAudio": true iff the mod uses plain audio
 * 	"files": {
 * 		file1: true // Replaces all of file1
 * 		file2: {
 * 		override: { "events": [1] } // Replaces only event 1
 * 			append: { "events": [1] } // Adds event 1 to the original map 
 * 		}
 * 		folder/* : true // Replaces all files from the mod folder
 * 	}
 * 	"folders": {
 * 		folder: true // Replaces all contents with the folder prefix
 * 	}
 * 	"plugins": [
 *		List of plugins as setup in the plugin.js
 *	]
 * }
 * 
 */

window.AuraMZ = window.AuraMZ || {};
AuraMZ.ModLoader = {};

$dataMods = [];

(($) => {
	$.MOD_FOLDER = "mods/";
	$.WILDCARD = "*";
	$.META = "Meta.json";

	// Contains the remapping information
	$.urlMap = {};

	// Contains the remapping information for full folders
	$.urlFolderMap = {};

	// Load a JSON file from the mod folder
	DataManager.loadModFile = function(name) {
		return new Promise((resolve, reject) => {
			const xhr = new XMLHttpRequest();
			xhr.open("GET", name);
			xhr.overrideMimeType("application/json");
			xhr.onload = () => (xhr.status < 400) ? resolve(JSON.parse(xhr.responseText)) : reject(this.onXhrError(name, name, name));
			xhr.onerror = () => reject(this.onXhrError(name, name, name));
			xhr.send();
		});
	};

	// Gets the url to the mod
	$.modUrl = function(modName) {
		return $.MOD_FOLDER + modName;
	};

	// Gets the path to the game
	$.gameDirectory = function() {
		const path = require("path");
		return path.dirname(process.mainModule.filename);
	};

	// Gets the list of files for the given game relative url
	$.modFiles = function(folderUrl) {
		const fs = require("fs");
		return fs.readdirSync($.gameDirectory() + "/" + folderUrl);
	};

	// Checks if a given folder is a mod folder (= Contains a Meta.json)
	$.isModFolder = function(folderName) {
		const fs = require("fs");
		return fs.existsSync($.gameDirectory() + "/" + $.MOD_FOLDER + folderName + "/" + $.META);
	};

	// Adds the .js suffix to plugin urls in the mods folder
	$.PluginManager_makeUrl = PluginManager.makeUrl;
	PluginManager.makeUrl = function(fileName) {
		if (fileName.startsWith($.MOD_FOLDER)) {
			return fileName + ".js";
		} else {
			return $.PluginManager_makeUrl.call(this, fileName);
		}
	};

	// List all mods in the mods folder (independently of whether they are enabled or not)
	$.listAllMods = function() {
		return $.modFiles($.MOD_FOLDER).filter(folder => $.isModFolder(folder));
	}

	// Load all mods
	$.loadMods = function() {
		if (StorageManager.isLocalMode()) {
			const fs = require("fs");
			const modsFolder = $.gameDirectory() + "/" + $.MOD_FOLDER;
			if (!fs.existsSync(modsFolder)) {
				fs.mkdirSync(modsFolder);
			}
			const mods = $.listAllMods().map((modFolder) => $.loadMod(modFolder));
			return Promise.all(mods).finally(() => {
				$dataMods.forEach(mod => $.applyMod(mod));
			});
		} else {
			return Promise.resolve();
		}
	};

	// Load a mod given its name
	$.loadMod = function(modName) {
		const modUrl = $.modUrl(modName);
		return DataManager.loadModFile(modUrl + "/" + $.META).then((mod) => {
			if (mod.enabled) {
				mod.name = modName;
				$dataMods.push(mod);
			}
		});
	}

	// Apply a loaded mod
	$.applyMod = function(mod) {
		if (mod.files) $.applyModFiles(mod);
		if (mod.folders) $.applyModFolders(mod);
		if (mod.plugins) $.applyModPlugins(mod);
	}

	// We always deal with urls without the potential encryption suffix
	$.stripEncryptionSuffix = function(fileName) {
		return fileName.endsWith("_") ? fileName.slice(0, -1) : fileName;
	};

	// Setup the map for URL redirection
	$.applyModFiles = function(mod) {
		const modUrl = $.modUrl(mod.name);
		for (const [fileName, value] of Object.entries(mod.files)) {
			if (value === true) {
				if (fileName.endsWith($.WILDCARD)) {
					const folderName = fileName.slice(0, -1);
					const modFiles = $.modFiles(modUrl + "/" + folderName)
						.map(modFile => $.stripEncryptionSuffix(modFile));
					modFiles.forEach(modFile => $.urlMap[folderName + modFile] = mod);
				} else {
					$.urlMap[fileName] = mod;
				}
			}
		}
	}

	// Setup the map for folder URL redirection
	$.applyModFolders = function(mod) {
		for (const [folderName, value] of Object.entries(mod.folders)) {
			if (value === true) {
				$.urlFolderMap[folderName] = mod;
			}
		}
	}

	// Load and setup mod plugins
	$.applyModPlugins = function(mod) {
		const modUrl = $.modUrl(mod.name);
		mod.plugins.forEach(plugin => plugin.name = modUrl + "/js/plugins/" + plugin.name);
		PluginManager.setup(mod.plugins);
	}

	// Inject the logic for loading mods
	$.DataManager_loadDatabase = DataManager.loadDatabase;
	DataManager.loadDatabase = function() {
		$.loadMods().finally(() => $.DataManager_loadDatabase.call(this));
	};

	// Apply the redirections from the url mappings
	$.updateURL = function(url) {
		let mod = $.urlMap[url];

		if (mod) {
			// File match
			url = $.modUrl(mod.name) + "/" + url;
		} else {
			const match = Object.entries($.urlFolderMap).find(([key]) => url.startsWith(key));
			if (match) {
				// Folder match
				const [folder, folderMod] = match;
				mod = folderMod;
				url = url.replace(folder, $.modUrl(mod.name) + "/" + folder);
			}
		}

		if (!mod) {
			return [null, url];
		}

		// Proceed recursively in case another mod redirects this url
		const [newMod, newUrl] = $.updateURL(url);
		return newMod ? [newMod, newUrl] : [mod, url];
	};

	// Inject url replacement from mods
	$.ImageManager_loadBitmapFromUrl = ImageManager.loadBitmapFromUrl;
	ImageManager.loadBitmapFromUrl = function(url) {
		const [mod, newUrl] = $.updateURL(url);
		const hasEncryptedImages = Utils.hasEncryptedImages();
		if (mod?.encryptedImages === false) {
			Utils._hasEncryptedImages = false;
		}
		const bitmap = $.ImageManager_loadBitmapFromUrl.call(this, newUrl);
		if (mod?.encryptedImages === false) {
			Utils._hasEncryptedImages = hasEncryptedImages;
		}
		return bitmap;
	};

	// Inject url replacement from mods
	$.WebAudio_initialize = WebAudio.prototype.initialize;
	WebAudio.prototype.initialize = function(url) {
		const [mod, newUrl] = $.updateURL(url);
		const hasEncryptedAudio = Utils.hasEncryptedAudio();
		if (mod?.encryptedAudio === false) {
			Utils._hasEncryptedAudio = false;
		}
		$.WebAudio_initialize.call(this, newUrl);
		if (mod?.encryptedAudio === false) {
			Utils._hasEncryptedAudio = hasEncryptedAudio;
		}
	};

	// Inject url replacement from mods
	DataManager.loadDataFile = function(name, src) {
		const xhr = new XMLHttpRequest();
		const url = "data/" + src;
		const [_, newUrl] = $.updateURL(url);
		window[name] = null;
		xhr.open("GET", newUrl);
		xhr.overrideMimeType("application/json");
		xhr.onload = () => this.onXhrLoad(xhr, name, src, newUrl);
		xhr.onerror = () => this.onXhrError(name, src, newUrl);
		xhr.send();
	};

	if (window["CTE"]) {
		$.CTE_load = CTE.load;
		CTE.load = function(localPath) {
			const [_, newUrl] = $.updateURL(localPath);
			return $.CTE_load.call(this, newUrl);
		}

		$.CTE_onLanguageLoad = CTE.onLanguageLoad;
		CTE.onLanguageLoad = function(xhr, url) {
			$.CTE_onLanguageLoad.call(this, xhr, url);
			$.applyModsToObject(CTE.languageObject, url);
		}
	}

	// Overwrite the default onLoad method to also pass the object url down
	$.DataManager_onXhrLoad = DataManager.onXhrLoad;
	DataManager.onXhrLoad = function(xhr, name, src, url) {
		if (xhr.status < 400) {
			window[name] = JSON.parse(xhr.responseText);
			this.onLoad(window[name], url);
		} else {
			this.onXhrError(name, src, url);
		}
	};

	// Applies mods to a loaded object (e.g. overrides and appends)
	$.applyModsToObject = function(object, url, loader, onLoad) {
		object.modApplications = [];
		object.modsApplied = false;

		for (const mod of $dataMods) {
			const value = mod.files[url];
			if (value != null && (value.override || value.append)) {
				object.modApplications.push(DataManager.loadModFile($.modUrl(mod.name) + "/" + url).then((modObject) => {
					// Make sure the meta data is being processed
					if (loader) onLoad.call(loader, modObject);

					if (value.override) $.applyOverrides(object, modObject, value.override);
					if (value.append) $.applyAppends(object, modObject, value.append);
				}));
			}
		}

		Promise.all(object.modApplications).finally(() => object.modsApplied = true);
		if (loader) onLoad.call(loader, object);
	}

	// Inject mod override and appends getting added into the object
	$.DataManager_onLoad = DataManager.onLoad;
	DataManager.onLoad = function(object, url) {
		$.applyModsToObject(object, url, this, $.DataManager_onLoad);
	};

	// Override the entries in the original object by the entries of the modded object
	$.applyOverrides = function(object, modObject, override) {
		if (Array.isArray(override) || override == $.WILDCARD) {
			$.applyModArray(object, modObject, override, $.applyArrayOverride);
		} else {
			for (const [property, propertyValue] of Object.entries(override)) {
				if (propertyValue === true) {
					// This is a basic property, simply assign the value
					const propertyPath = property.split('.');
					const accessModObject = $.accessProperty(modObject, propertyPath);
					const accessProperty = propertyPath.pop();
					const accessObject = $.accessProperty(object, propertyPath);
					accessObject[accessProperty] = accessModObject;
				} else if (propertyValue === false) {
					// Skip this override
					continue;
				} else if ($.isArrayProperty(property)) {
					// This is a list of IDs, select the matching objects and drill deeper	
					$.processArrayProperty(object, modObject, property, propertyValue, $.applyOverrides);
				} else {
					// This is a sub property, drill deeper into the objects
					const propertyPath = property.split('.');
					$.applyOverrides($.accessProperty(object, propertyPath), $.accessProperty(modObject, propertyPath), propertyValue);
				}
			}
		}
	}

	// Append the modded objects into the object list of the original
	$.applyAppends = function(object, modObject, append) {
		if (Array.isArray(append) || append == $.WILDCARD) {
			// This is an array, so we just append at the end of the array
			$.applyModArray(object, modObject, append, $.applyArrayAppend);
		} else {
			// This is an object, so we have to iterate over all declared properties
			for (const [property, objects] of Object.entries(append)) {
				if ($.isArrayProperty(property)) {
					// This is a list of IDs, select the matching objects and drill deeper	
					$.processArrayProperty(object, modObject, property, objects, $.applyAppends);
				} else {
					// This is a regular property, drill deeper
					const propertyPath = property.split('.');
					$.applyAppends($.accessProperty(object, propertyPath), $.accessProperty(modObject, propertyPath), objects);
				}
			}
		}
	}

	// Apply the operation to the target object using the mod objects
	$.applyModArray = function(object, modObject, objectIds, operation) {
		objectIds = objectIds == $.WILDCARD
			? modObject.filter(obj => obj).map((obj, index) => obj?.id ? obj.id : index)
			: objectIds
		objectIds.forEach(objectId => operation(object, modObject, objectId));
	}

	// Perform an override operation
	$.applyArrayOverride = function(object, modObject, objectId) {
		object[objectId] = $.getObjectById(modObject, objectId);
	}

	// Perform an append operation
	$.applyArrayAppend = function(object, modObject, objectId) {
		const appendObject = $.getObjectById(modObject, objectId);
		object.push(appendObject);
		appendObject.id = object.length - 1;
	}

	// Get an object by ID, if the object has no id, use the array index
	$.getObjectById = function(array, id) {
		return array.find((obj, index) => obj?.id ? obj.id == id : index == id);
	}

	// Checks if the current key in the mod is a list of affected items
	$.isArrayProperty = function(property) {
		return (property.startsWith("[") && property.endsWith("]")) || property == $.WILDCARD;
	}

	// Processes properties of the form { "[id1, id2, ...]" : ... } by either
	// parsing the key into an array or resolving the wildcard if it contains one
	$.processArrayProperty = function(object, modObject, property, objects, operation) {
		const objectIds = property == $.WILDCARD
			? modObject.filter(obj => obj).map(obj => obj.id)
			: JSON.parse(property);

		objectIds.forEach(objectId => {
			operation($.getObjectById(object, objectId), $.getObjectById(modObject, objectId), objects);
		});
	}

	// Walks the property path down the given object
	$.accessProperty = function(object, propertyPath) {
		propertyPath.forEach(segment => object = object[segment]);
		return object;
	}

	// Inject logic to ensure that all mods have been applied before progressing the game
	$.DataManager_isMapLoaded = DataManager.isMapLoaded;
	DataManager.isMapLoaded = function() {
		return $.DataManager_isMapLoaded.call(this) && ($dataMap.modsApplied == undefined || $dataMap.modsApplied);
	};
})(AuraMZ.ModLoader);