//=============================================================================
// RPG Maker MZ - Pathfinding
// ----------------------------------------------------------------------------
// (C)2021 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 2025/02/12
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Optimized pathfinding algorithm
 * @author aura-dev
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help pathfinding.js
 *
 * Improves the standard pathfinding algorithm.
 * 
 * After:
 * - detectors.js
 * 
 * Added tags:
 * <forced_collision>
 * - Forces events with the through flag to check for tile collision.
 * Useful for small background events roaming the map without entering inacessible tiles or blocking other characters.
 *
 * @param searchLimit
 * @text Search Limit
 * @type number
 * @desc Search limit for the A* pathfinding algorithm. If exceeded, will pick most promising node within search limit.
 * @default 75
 */

window.AuraMZ = window.AuraMZ || {};
AuraMZ.Pathfinding = {};

(($) => {
	const PLUGIN_ID = "pathfinding";
	const params = PluginManager.parameters(PLUGIN_ID);
	$.searchLimit = parseInt(params["searchLimit"]);
	
	const TAG_FORCED_COLLISION = "<forced_collision>"

	// Change the logic so that if an event is near the player it will randomly choose between
	// naively stepping towards the character and using a shortest route movement.
	//
	// To prevent far events from getting stuck on what they perceive to be the shortest
	// route because they are outside the search range, we keep the original algorithm
	// for injecting random movements for events that are not near the player
	Game_Event.prototype.moveTypeTowardPlayer = function() {
		if (this.isNearThePlayer()) {
			this.moveTowardCharacterPathfinding(0);
		} else {
			this.moveTowardCharacterApprox(0);
		}
	};

	// Perform a far-range movement (or if there is no path to the player)
	// so that enemies aren't still in that case
	Game_Character.prototype.moveTowardCharacterApprox = function(id) {
		switch (Math.randomInt(6)) {
			case 0:
			case 1:
			case 2:
				const char = id == 0 ? $gamePlayer : $gameMap.event(id);
				this.moveTowardCharacter(char);
				break;
			case 3:
			case 4:
				this.moveRandom();
				break;
			case 5:
				this.moveForward();
				break;
		}
	}

	// A move toward player using pathfinding
	Game_Character.prototype.moveTowardCharacterPathfinding = function(id) {
		const char = id == 0 ? $gamePlayer : $gameMap.event(id);
		direction = this.findDirectionTo(char.x, char.y);
		if (direction > 0) {
			this.moveStraight(direction);
			const deltaDirectionToCharacter = (direction == 4 || direction == 6)
				? this.deltaXFrom(char.x) : this.deltaYFrom(char.y);
			if (!this.isMovementSucceeded() && deltaDirectionToCharacter > 1) {
				this.moveTowardCharacterApprox(id);
			}
		} else {
			this.moveTowardCharacterApprox(id);
		}
	};

	// Include the searchLimit for pathfinding in the check of 
	// an event is considered near the player or too far to
	// use pathfinding for a move towards action
	Game_Event.prototype.isNearThePlayer = function() {
		const searchLimit = this.searchLimit();
		const sx = Math.abs(this.deltaXFrom($gamePlayer.x));
		const sy = Math.abs(this.deltaYFrom($gamePlayer.y));
		return sx + sy < Math.max(20, searchLimit / 2);
	};

	Game_Character.prototype.findDirectionTo = function(goalX, goalY) {
		const searchLimit = this.searchLimit();
		const mapWidth = $gameMap.width();
		const nodeList = [];
		const openList = [];
		// Use a list instead of a set for the closed list
		// This makes the check if a node is in the closed list
		// O(log n) instead of O(n) when using a regular list 
		const closedList = new Set();
		const start = {};
		let best = start;

		if (this.x === goalX && this.y === goalY) {
			return 0;
		}

		start.parent = null;
		start.x = this.x;
		start.y = this.y;
		start.g = 0;
		start.f = $gameMap.distance(start.x, start.y, goalX, goalY);
		nodeList.push(start);
		openList.push(start.y * mapWidth + start.x);

		while (nodeList.length > 0) {
			let bestIndex = 0;
			for (let i = 0; i < nodeList.length; i++) {
				if (nodeList[i].f < nodeList[bestIndex].f) {
					bestIndex = i;
				}
			}

			const current = nodeList[bestIndex];
			const x1 = current.x;
			const y1 = current.y;
			const pos1 = y1 * mapWidth + x1;
			const g1 = current.g;

			nodeList.splice(bestIndex, 1);
			openList.splice(openList.indexOf(pos1), 1);
			closedList.add(pos1);

			if (current.x === goalX && current.y === goalY) {
				best = current;
				break;
			}

			if (g1 >= searchLimit) {
				continue;
			}

			for (let j = 0; j < 4; j++) {
				const direction = 2 + j * 2;
				const x2 = $gameMap.roundXWithDirection(x1, direction);
				const y2 = $gameMap.roundYWithDirection(y1, direction);
				const pos2 = y2 * mapWidth + x2;

				if (closedList.has(pos2)) {
					continue;
				}
				if (!this.canPass(x1, y1, direction)) {
					continue;
				}

				const g2 = g1 + 1;
				const index2 = openList.indexOf(pos2);

				if (index2 < 0 || g2 < nodeList[index2].g) {
					let neighbor = {};
					if (index2 >= 0) {
						neighbor = nodeList[index2];
					} else {
						nodeList.push(neighbor);
						openList.push(pos2);
					}
					neighbor.parent = current;
					neighbor.x = x2;
					neighbor.y = y2;
					neighbor.g = g2;
					neighbor.f = g2 + $gameMap.distance(x2, y2, goalX, goalY);
					if (!best || neighbor.f - neighbor.g < best.f - best.g) {
						best = neighbor;
					}
				}
			}
		}

		let node = best;
		while (node.parent && node.parent !== start) {
			node = node.parent;
		}

		const deltaX1 = $gameMap.deltaX(node.x, start.x);
		const deltaY1 = $gameMap.deltaY(node.y, start.y);
		if (deltaY1 > 0) {
			return 2;
		} else if (deltaX1 < 0) {
			return 4;
		} else if (deltaX1 > 0) {
			return 6;
		} else if (deltaY1 < 0) {
			return 8;
		}

		const deltaX2 = this.deltaXFrom(goalX);
		const deltaY2 = this.deltaYFrom(goalY);
		if (Math.abs(deltaX2) > Math.abs(deltaY2)) {
			return deltaX2 > 0 ? 4 : 6;
		} else if (deltaY2 !== 0) {
			return deltaY2 > 0 ? 8 : 2;
		}

		return 0;
	};

	// Increase search limit for pathfinding
	Game_Character.prototype.searchLimit = function() {
		return $.searchLimit;
	};

	// Prevent through events from passing through tile collision if they have the forced_collision tag
	const _Game_CharacterBase_canPass = Game_CharacterBase.prototype.canPass;
	Game_CharacterBase.prototype.canPass = function(x, y, d) {
		if (this.event?.().note.contains(TAG_FORCED_COLLISION) && this.isThrough()) {
			const x2 = $gameMap.roundXWithDirection(x, d);
			const y2 = $gameMap.roundYWithDirection(y, d);
			return $gameMap.isValid(x2, y2) && this.isMapPassable(x, y, d);
		} else {
			return _Game_CharacterBase_canPass.call(this, x, y, d);
		}
	};

	// Utility function to check if a character can only move backwards
	Game_Character.prototype.canOnlyMoveBackwards = function() {
		const d = this.direction();
		const oppositeD = this.reverseDir(d);
		const directions = [2, 4, 6, 8].remove(oppositeD);
		return directions.every(dir => !this.canPass(this.x, this.y, dir));
	}

	if (AuraMZ.Detectors) {
		// Prevent backwards direction for enemies with the detector tag
		const _Game_Character_moveRandom = Game_Character.prototype.moveRandom;
		Game_Character.prototype.moveRandom = function() {
			const detectorTag = this.getDetectorTag();
			if (this.canChase(detectorTag)) {
				let d;
				const oppositeD = this.reverseDir(this.direction());
				do {
					if (this.canOnlyMoveBackwards()) {
						d = oppositeD;
						break;
					} else {
						d = 2 + Math.randomInt(4) * 2;
					}
				} while (d === oppositeD);
				if (this.canPass(this.x, this.y, d)) {
					this.moveStraight(d);
				}
			} else {
				_Game_Character_moveRandom.call(this);
			}
		};
	}
})(AuraMZ.Pathfinding);