/*:
* @plugindesc Loads script files from a languages folder to dynamically use text and more for the chosen language.
* @author vgperson
*
* @param Important
* @text Important Settings
* @desc Important settings for configuring supported languages.
*
* @param DefaultLanguageCodes
* @text Default Language Codes
* @parent Important
* @type text[]
* @desc Default list of language codes. This will be the order they appear in the options menu. Overriden if a Languages.json is provided in the languages folder.
* @default ["JP","EN"]
*
* @param BaseLanguage
* @text Base Language Code
* @parent Important
* @type text
* @desc Base game language; this language will use game data rather than language scripts. Must match a Language Code.
* @default JP
*
* @param DefaultLanguage
* @text Default Language Setting
* @parent Important
* @type text
* @desc Default language setting, used on first launch. Must match a Language Code. Base Language used if blank.
* @default EN
*
* @param BaseLanguageDisplayName
* @text Language Name (Base)
* @parent Important
* @type text
* @desc Base language name for display in options menu. (Defined in Extra section for other languages.)
* @default 日本語
*
* @param BaseLanguageOptionName
* @text Language Option (Base)
* @parent Important
* @type text
* @desc Name of the "Language" option in options menu in base language. (Defined in Extra section for other languages.)
* @default 言語
*
* @param PluginDatabases
* @text Plugin Databases
* @parent Important
* @type text[]
* @desc An array of databases added by plugins that should be considered for translation.
* @default []
*
* @param BaseMessageSpeed
* @text Message Speed (Base)
* @parent Important
* @type combo
* @option 0.125x
* @option 0.25x
* @option 0.5x
* @option 0.66x
* @option 0.75x
* @option 0.8x
* @option 0.9x
* @option 1x
* @option 1.1x
* @option 1.125x
* @option 1.16x
* @option 1.2x
* @option 1.25x
* @option 1.33x
* @option 1.5x
* @option 1.66x
* @option 2x
* @desc Message speed for base language. (Defined in Extra section for other languages.) Default is 1x.
* @default 1x
*
* @param Encryption
* @text Encryption Settings
* @desc Settings for using encrypted CTE language files instead of TXT files.
*
* @param LoadCTEFiles
* @text Load From CTE Files
* @parent Encryption
* @type select
* @option Load From TXTs
* @value false
* @option Load From CTEs
* @value true
* @option Load From CTEs, TXTs For Playtest
* @value txtplaytest
* @desc Whether to load encrypted CTE language files rather than plain TXT scripts. Use to obfuscate for public releases.
* @default false
*
* @param GenerateCTEFiles
* @text Generate CTE Files
* @parent Encryption
* @type boolean
* @desc Whether to, whenever a ".json" language script is loaded, generate the equivalent encrypted ".cte" language file.
* @default false
*
* @param Launch
* @text Launch Actions
* @desc Settings for whether to perform certain actions at launch in playtest mode.
*
* @param ExportBaseOnLaunch
* @text Export Base Text
* @parent Launch
* @type boolean
* @desc Whether to, on launch, export base language text from data. Be careful not to overwrite files unwantedly.
* @default false
*
* @param BaseExportFilename
* @text Base Export Filename
* @parent ExportBaseOnLaunch
* @type text
* @desc File/folder name to export base language text to. If left blank, it will use the Base Language Code.
* @default
*
* @param AllInOneFile
* @text Export to Single File
* @parent ExportBaseOnLaunch
* @type boolean
* @desc Whether to export language files as a single all-in-one .txt file. Turn off to export as a folder using sub-settings.
* @default true
*
* @param AllInOneMaps
* @text All-In-One Maps
* @parent ExportBaseOnLaunch
* @parent AllInOneFile
* @type boolean
* @desc Whether to export all maps to a single "AllMaps.txt" rather than individual map files.
* @default false
*
* @param AllInOneCommons
* @text All-In-One Commons
* @parent ExportBaseOnLaunch
* @parent AllInOneFile
* @type boolean
* @desc Whether to export all common events to a single "AllCommons.txt" rather than individual event files.
* @default true
*
* @param AllInOneTroops
* @text All-In-One Troops
* @parent ExportBaseOnLaunch
* @parent AllInOneFile
* @type boolean
* @desc Whether to export all troops to a single "AllTroops.txt" rather than individual troop files.
* @default true
*
* @param AllInOneDatabase
* @text All-In-One Database
* @parent ExportBaseOnLaunch
* @parent AllInOneFile
* @type boolean
* @desc Whether to export all other databases to a single "AllDatabases.txt" rather than individual tab files.
* @default true
*
* @param NotetagsToExport
* @text Notetags to Export
* @parent Launch
* @type string[]
* @desc An array of tags. If a skill or item has a tag in the list, the note will be exported.
* @default []
*
* @param PluginParamsToExport
* @text Plugin Parameters To Export
* @parent Launch
* @type text[]
* @desc An array of parameter names / suffixes. If a plugin parameter ends in it, will be exported.
* @default ["_cte"]
*
* @param ExportTableOnLaunch
* @text Export Reference Table
* @parent Launch
* @type boolean
* @desc Whether to, on launch, export a CSV file comparing all languages. Useful for checking mismatched text.
* @default false
*
* @param ExportIdenticalOnLaunch
* @text Log Identical Strings
* @parent Launch
* @type boolean
* @desc Whether to, on launch, log to IdenticalWarnings.txt with text that is identical in multiple languages.
* @default false
*
* @param Other
* @text Other Settings
* @desc Miscellaneous settings.
*
* @param MakeImageList
* @text Make Image List
* @parent Other
* @type boolean
* @desc During playtests, auto-generate TranslatedImages.txt for auto-loading of translated images outside of NW.js.
* @default true
*
* @param LoadAnyLanguageFiles
* @text Load Any Language Files
* @parent Other
* @type select
* @option Only Load Listed Languages
* @value false
* @option Load Listed + Additional Languages
* @value true
* @option Load Listed Always, Others If Not Playtest
* @value notinplaytest
* @desc Whether to load any non-listed languages found in languages folder, allowing users to easily put in new languages.
* @default false
*
* @param LockLanguage
* @text Lock to Single Language
* @parent Other
* @type boolean
* @desc Forces game to use Default Language Setting, and removes the language setting from the Options menu.
* @default false
*
* @param ForcePlaytestMode
* @text Force Playtest Mode
* @parent Other
* @type boolean
* @desc Act like it's playtest mode even in a release version. (File-related functions still require it be NW.js.)
* @default false
*
* @param ReloadKey
* @text Script Reload Key
* @parent Other
* @type select
* @option Disabled
* @value 0
* @option F1
* @value 112
* @option F2
* @value 113
* @option F3
* @value 114
* @option F4
* @value 115
* @option F5
* @value 116
* @option F6
* @value 117
* @option F7
* @value 118
* @option F8
* @value 119
* @option F9
* @value 120
* @option F10
* @value 121
* @option F11
* @value 122
* @option F12
* @value 123
* @desc Debug key during playtests to reload all language scripts. Hold to show actions menu. Default F10 (keycode 121).
* @default 121
*
* @param MessageDebugKey
* @text Message Debugger Key
* @parent Other
* @type select
* @option Disabled
* @value 0
* @option F1
* @value 112
* @option F2
* @value 113
* @option F3
* @value 114
* @option F4
* @value 115
* @option F5
* @value 116
* @option F6
* @value 117
* @option F7
* @value 118
* @option F8
* @value 119
* @option F9
* @value 120
* @option F10
* @value 121
* @option F11
* @value 122
* @option F12
* @value 123
* @desc Debug key during playtests to open message previewer. Press again in previewer to search. Default F11 (keycode 122).
* @default 122
*
* @param LogSetting
* @text Console Log Detail
* @parent Other
* @type select
* @option None
* @value 0
* @option Load/Export Messages
* @value 1
* @option Load/Export + Missing Translations
* @value 2
* @desc How much to log to the console. Level 1: Language file loading & exporting. Level 2: Include translation warnings.
* @default 2
*
* @help
* **********************************************************************
* *                vgperson's Custom Translation Engine                *
* **********************************************************************
*
* This plugin loads translation data from files in a "languages" folder,
* and will then automatically use appropriate text for the current language.
*
* Besides text from message/choice commands and database text,
* language files can also hold arbitrary user-added "extra" data
* to be referred to by plugin scripts or script commands.
*
* There are also a wide variety of other functions
* that can help assist with translation and checking for mistakes.
*
* For basic instructions on how to set things up and
* what most of the settings are for, read the Setup Procedure section.
* For details on all the different functions, read Feature Explanations.
*
* ***** Usage Notes *****
* Anyone is free to use this plugin for their game/translation projects
* and make modifications as necessary.
* Explicitly giving credit to me (vgperson) is appreciated.
* Please keep these usage notes and the Credits section intact,
* and be mindful about distributing language files that aren't yours.
*
*
* ======================================================================
*                            Setup Procedure
* ======================================================================
*
* 1. Add and enable this plugin in the Plugin Manager.
*
* Where you place it in the plugin list should usually only matter
* if you make plugins refer to its functions (i.e. CTE.getString()).
*
* If you put it at the very top, CTE will be initialized before anything,
* so you can safely directly edit plugins to refer to CTE functions,
* but may have issues if you add CTE code that refers to other plugins.
*
* Alternatively, you can place it at the bottom of the list,
* and override the definitions of plugin functions as necessary
* within CustomTranslationEngine.js itself.
*
* --------------------------------------------------
*
* 2. Set up the Language Codes list.
*
* These codes represent all of the supported languages for the game,
* in the same order you want them displayed in the Options menu.
*
* You should be able to add to or rearrange these at any time without issue.
*
* --------------------------------------------------
*
* 3. Set the Base Language Code and Default Language Setting.
*
* The Base Language Code is the language of the base game.
* When a player is playing in this language, the base game data will be used.
*
* The Default Language Setting is the language the game initially launches in.
* This is overridden by saved user configuration data (i.e. config.rpgsave),
* which is made once the player goes into the Options menu.
* If left blank, the default setting will be the Base Language.
*
* You can have language settings from a launcher such as Steam
* override the Default Language Setting if configured properly.
* (See Customizable Handlers section.)
*
* --------------------------------------------------
*
* 4. Set up core language settings for the base language.
*
* ***** Language Name (Base) *****
* Name of the base language as shown in the Options menu.
*
* ***** Language Option (Base) *****
* Text for the "Language" option in the Options menu in the base language.
*
* ***** Message Speed (Base) *****
* Message typing speed in base language. Default 1x (1 character per frame).
* If you're using another plugin that modifies message speed,
* speeds other than 1x may override its functionality, so be careful.
*
* Note that for languages other than the base language,
* these settings are defined in the Extra section of their language files.
* (See Core Language Settings section.)
*
* --------------------------------------------------
*
* 5. To start translating, enable Launch Actions > Export Base Text,
* then run the game to generate base language data.
*
* This will put text data for the Base Language in the "languages" folder,
* which can then be renamed to a different language code and translated.
* (Launch actions like this only run in playtest mode and shouldn't occur
* in releases, but you might want to disable them when not needed.)
*
* The exported data is named after the Base Language Code by default,
* but will use the Base Export Filename if one is specified.
*
* You can also export base game text via the Export Actions menu,
* accessed by holding down the Script Reload key (defaulf F10) in-game.
*
* ***** "Export to Single File" Settings *****
* By default, language text is exported to a single .txt file.
* If desired, you can have it split up across multiple files in a subfolder.
*
* If "Export to Single File" is on, all text will go in a single "XX.txt"
* (where XX is the Base Language Code or Base Export Filename).
* If "Export to Single File" is off, an "XX" folder will be made,
* which will contain multiple files based on the other settings.
*
* Having "All-In-One Maps" on will put all map data into "AllMaps.txt",
* whereas having it off will generate an individual file for every map.
* The same pattern applies for Common Events, Troops, and Databases,
*
* NOTE: These settings do NOT affect how language files are loaded.
* Rather, loading always prioritizes "all-in-one" files first,
* then looks for individual files if those do not exist.
*
* For example, if a single "XX.txt" exists for the language code,
* it will be loaded and any other files that may exist will be ignored.
* If it does not, however, the plugin will check for each all-in-one file
* (AllMaps.txt, AllCommons.txt, AllTroops.txt, AllDatabases.txt),
* and if one does not exist, looks for individual files of that type.
*
* ***** A Note On Extraneous Language Files *****
* For public releases, you only need to include
* language files for languages other than the Base Language.
* The Base Language just uses the text that exists in the game data,
* so it doesn't need its own external language files.
*
* If set up to load CTE files, you only need the ".cte"s for each language,
* and TXT language files/subfolders can be omitted for release.
* If you're not using CTEs, only the TXTs for each language are necessary.
*
* A TranslatedImages.txt file is only necessary if you're using
* alternate image loading and intend for the game to be played in-browser.
* ReferenceTable.csv and IdenticalWarnings.txt files
* are both purely for reference and can be omitted.
*
* Leaving extra files in won't cause any actual issues, but it may be a concern
* if you don't want people looking through plain text or development data.
*
* --------------------------------------------------
*
* 6. (Optional) Configure other settings as you wish.
*
* ***** Encryption Settings (Load From / Generate CTE Files) *****
* While you can use TXT-format language files directly,
* when actually publishing the game, you may want that data
* to be a compressed, encrypted file instead of plain text.
* This can be accomplished by generating CTE files.
*
* If you enable "Generate CTE Files," then in playtest mode,
* an encrypted CTE file will be generated every time
* TXT language files are loaded in (to keep it up to date).
* Then, for release, you can enable "Load From CTE Files"
* to load language data from CTE files instead of TXTs.
*
* Using this option, you can include just the CTE files in your release.
* That said, it's recommended you hold onto the original TXT files.
* While it is possible to recreate the TXT files from a CTE file
* (by loading the CTE and using Export Reformatted Language Text),
* any comments that were included in the TXTs will be lost.
*
* You can export the latest CTEs anytime from the Export Actions menu,
* accessed by holding down the Script Reload key in-game.
*
* ***** Make Image List *****
* Makes the game auto-generate the file TranslatedImages.txt,
* which lists all image files that have translated versions in each language.
* This file is only needed if the game is intended to be played in-browser.
* (See Alternate Image Loading.)
*
* ***** Load Any Language Files *****
* This setting helps allow players to easily add languages beyond
* what is specified in the Language Codes list.
*
* If enabled, after loading "built-in" languages, the plugin will check for
* any other potential language files/subfolders in the languages folder.
* If they seem valid, they will be loaded and appended to the language list.
*
* A CTE file is considered valid as long as it parses without issues.
* A TXT file is considered valid if it has a command definition or the like.
* A subfolder is considered valid if it has at least one valid TXT file in it
* that uses a recognized name. (Map001, AllDatabases, Actors, etc.)
*
* Be careful about leaving excess TXT/CTE files in the language folder,
* as it could result in unwanted language options showing up.
* In particular, reformatted language files are liable to be loaded
* as new languages (and even generate their own CTEs, if enabled).
* If this proves annoying, the "Others If Not Playtest" option will
* disable the loading of additional languages while in playtest mode.
*
* ***** Lock to Single Language *****
* If you want to release the game in one specific language and not offer
* the option to switch languages, this setting will force the language to
* the Default Language Setting and not add the in-game language option.
*
* This can be used for any language code, including the Base Language.
* It can be helpful if you want to release a base-language-only version,
* but are unable to just remove the CTE plugin since other code refers to it.
*
* ***** Force Playtest Mode *****
* If, for whatever reason, you want this plugin's "playtest-only" functions
* to be usable even outside of playtesting, you can enable this setting
* to make it always behave like a playtest.
* Be sure to turn this off for releases if you don't want it enabled.
*
* Note that this simply overrides the playtest checks in THIS plugin,
* and will not enable other playtest features such as the debug console.
* Also, any file-related functions (like export functions) still require
* that you be in NW.js (AKA a standalone executable).
*
* ***** Script Reload Key *****
* A debug key for reloading all language scripts mid-game.
* Only usable in playtest mode. Default is F10 (keycode 121).
*
* Pressing the Reload Key while a message box is up
* will re-display that message using the latest data.
*
* If you press the Reload Key with Load From CTE Files enabled,
* it will always load TXT files, then re-export them as CTE files.
*
* Holding the Reload Key down in-game (while not in a scene) will display
* the Export Actions menu, which lets you perform the following:
* - Export text for the Base Language (same as launch action)
* - Export the Reference Table (same as launch action)
* - Log identical text between languages (same as launch action)
* - Generate .cte files from the latest .txt data (done on launch if enabled)
* - Export "reformatted" .txt language files for current language
* - Open the Message Typer for immediate message previewing
*
* ***** Message Debugger Key *****
* A debug key that opens a previewer for all messages in current language.
* Only usable in playtest mode. Default is F11 (keycode 122).
* Press the key again while the message debugger is open to search.
*
* ***** Console Log Detail *****
* How much console logging the plugin should do.
*
* None: Nothing will be logged to the console.
*
* Level 1 (Load/Export Messages):
* Logs info about files being loaded or exported.
*
* Level 2 (Load/Export + Missing Translations):
* In addition to Level 1 log messages,
* warns about missing or inconsistent translations.
*
*
*
* ======================================================================
*                          Feature Explanations
* ======================================================================
*
* ******************** FEATURE SECTION OVERVIEW ********************
*
* --- Command Text Replacement
* Info on the basic functionality of replacing text in commands.
*
* --- Database Text Replacement
* Similar to the above, but with minor differences in how databases work.
*
* --- Notes On Special Lookups
* Special cases where translations must be accompanied by source strings.
*
* --- Translating Strings in Variables
* Localizing strings in variables without directly altering them.
*
* --- Translating Message Speaker Names (MZ)
* Localizing message speaker names (RPG Maker MZ-only feature) all in one place.
*
* --- Core Language Settings
* Important settings for each language, included in the Extra database section.
*
* --- Defining Custom Values
* Making your own language-based values for use by plugins and such.
*
* --- Alternate Image Loading
* Setting things up to auto-load translated images for the current language.
*
* --- Language-Conditional Pages
* Making event pages that only appear/run in certain languages.
*
* --- Command Text Conditions
* Adding new conditions for text without having to manually split commands.
*
* --- Language Comparison Table
* A reference resource for comparing languages side by side.
*
* --- Logging Identical Strings
* A function to help locate text that may not be translated yet.
*
* --- Ignoring Text Issues
* If text need not be translated, indicate where you want to ignore warnings.
*
* --- Message Debugger Menu
* A testing menu for previewing all translated messages in-game.
*
* --- Message Typer Window
* A debug window for previewing arbitrary messages in-game.
*
* --- Reformatting Language Files
* An export function to rewrite language scripts into a standard format.
*
* --- Customizable Handlers
* Guides for how to customize this plugin to tailor it to a specific game.
*
*
*
* ----------------------------------------------------------------------
*  Command Text Replacement
* ----------------------------------------------------------------------
*
* Maps, Common Events, and Troop pages are structured as
* a list of all commands which contain text to be translated.
*
* Current map/file is set by markers in the format "=MapX=" or "=File=".
* Current event is set by markers in the format "*EventX*".
* Current page is set by markers in the format "-PageX-".
*
* (*CommonX* is used instead of *EventX* for Common Events
* (though either works), and *TroopX* is used within Troops.
* Also, common events don't have multiple pages.
* If using individual files instead of a single language file,
* some initial values are assumed based on the file; for instance,
* Common0001.txt has an implicit "=CommonEvents=" and "*Common1*".)
*
* Within a section, commands are given in a format like this:
* #Message#
* First line of message
* Second line of message
* ##
*
* The header, #Message#, indicates the type of command.
* The lines that follow make up the text content of the command.
* The terminator ## ends the command.
*
* Commands of the same type in the same "domain" (i.e. map page)
* go into an ordered list of command translations. This becomes the order
* in which they match to commands of that type in-game.
*
* Example:
* --------------------------------------------------
* [Event commands]
* Show Text: Hello!
* Wait: 10 frames
* Show Choice: Yes / No
* (arbitrary amount of non-Show Text commands)
* Show Text: Hello again!
* --------------------------------------------------
* [languages/EN.txt]
* #Message#
* Hello!
* ##
*
* #Choice#
* Yes
* No
* ##
*
* #Message#
* Hello again!
* ##
* --------------------------------------------------
*
* In this way, even if there are changes to the surrounding commands,
* you only have to change the language scripts
* if the order of text-containing commands changes,
* or if the map numbers, event IDs, or page numbers change.
*
* Also, a variety of contextual information is included on export,
* such as event names, coordinates, and choice branch names.
* This information is ignored when scripts are read;
* it is not critical to functionality and affects nothing.
*
* Similarly, outside of command content, as long as section markers are intact,
* you can add any text you wish as a "comment" and it will be ignored.
* (Example: "#Message# comment text" still reads as a Message command header.)
*
* An additional feature: If you want to skip a Message textbox entirely,
* define its translation as the following string:
* <<REMOVE>>
*
*
* ----------------------------------------------------------------------
*  Database Text Replacement
* ----------------------------------------------------------------------
*
* Database text uses a similar format to command text.
* The start of each database is marked by "=DatabaseName=",
* and database entries (if it's made up of entries) are marked by "*EntryX*".
*
* Fields within each entry, like "#Name#", are formatted just like commands.
* The only difference is that there's only ever one of each field per entry.
*
*
* ----------------------------------------------------------------------
*  Notes On Special Lookups
* ----------------------------------------------------------------------
*
* In cases where text is stored within the save file,
* this plugin opts to only ever store text from the Base Language,
* but localize it to the current language when displaying it to the player.
* In order to create original-to-translation "lookups" to accomplish this,
* these cases are formatted a bit differently in the scripts.
*
* The DisplayName field for maps uses a special format
* which includes the original text ("#DisplayName:ORIGINAL#"),
* to help support plugins that show map names on the file menu.
* *Do not alter the original text* - it will break that functionality!
*
* NameChange, NicknameChange, and ProfileChange commands work similarly;
* again, *altering the original text will break functionality.*
* Note that values from the fields in the Actor database also go into
* the name/nickname/profile lookups, but need not specify original text
* since they can automatically map to the base data.
*
*
* --------------------------------------------------------------------------
*  Translating Strings in Variables
* --------------------------------------------------------------------------
*
* Strings in variables can also be translated with a lookup-based approach.
* In the "Variables" section, you can define strings and their translations,
* and those strings will be converted for display purposes only.
*
* The plugin will attempt to locate as many string variable assignments
* as it can for export, but you may have to add some manually.
*
* The format for a definition is a header (either ## or #[options]#),
* followed by a line for the original, followed by a line for the translation.
*
* Example:
* ##
* Original text
* Translated text
*
* This would make it so that whenever a variable contains the exact,
* case-sensitive string "Original text", it will be displayed as
* "Translated text" in the language where this rule is defined.
* This should not affect any inner workings, and since the data keeps
* the original string, no problems should occur from switching languages.
*
* If you only want a translation to be applied to specific variables
* (for instance, a case where you want the same text to be
* translated differently for different variables), you can specify
* one or more variables to target (use commas for a list) in the header.
*
* Example:
* #V1,V2#
* Original text
* Translated text
*
* The above rule would only be applied to Variable 1 or Variable 2.
* Specifically-targeted rules take priority over general ones.
*
* You can also use regular expressions if you put "regex" in the header.
* Regular expressions must match the entire string, not just part of it.
* Captured groups in parentheses can be used in the translated text
* by adding zero-based numbers in curly braces, i.e. {0} {1} {2}.
*
* Examples:
* #regex#
* First (.*), then (.*)
* {0} is first, {1} is second
*
* #V1,regex#
* This is (.*), that is (.*)
* That is {1}, this is {0}
*
* The latter rule would swap the order of the captured words.
*
* ***** Notes On Extending Functionality *****
* This plugin only applies variable translation rules
* when converting escape characters for display in messages.
* So if you use plugins that display string variables
* without going through the convertEscapeCharacters() function,
* you must make modifications to localize those strings.
*
* If a plugin refers to the function $gameVariables.value(number),
* you can force it to be localized simply by adding a "true" argument
* (i.e. "$gameVariables.value(1, true)").
*
* Alternatively, if you set "CTE.localizeVariableOverride" to true,
* any calls to the $gameVariables.value function will be localized.
* But you should be careful to only use this in a temporary manner,
* setting it back to false as soon as the relevant code finishes;
* if you don't, it may cause issues like localized strings being
* written to save data and no longer changing to match language.
*
*
* --------------------------------------------------------------------------
*  Translating Message Speaker Names (MZ)
* --------------------------------------------------------------------------
*
* The "Speakers" database section in the language scripts will only be included
* if the game uses the speaker name functionality added in RPG Maker MZ.
*
* When exporting base text, the plugin will create a list of all unique names
* it finds across all messages to define translations for.
*
* The format for a speaker name definition is a header (##),
* followed by a line for the original, followed by a line for the translation.
*
* Example:
* ##
* Original name
* Translated name
*
*
* --------------------------------------------------------------------------
*  Core Language Settings
* --------------------------------------------------------------------------
*
* The "Extra" database section in the language scripts contains
* some miscellaneous settings that each language should define.
* Note that the Base Language's definitions for these are instead
* given through the plugin settings.
*
* ***** LanguageDisplayName *****
* Name of the language as shown in the Options menu.
* If not defined, the language code itself will be used.
*
* ***** LanguageOptionName *****
* Text for the "Language" option in the Options menu.
* If not defined, the Base Language version will be used.
*
* ***** DefaultFont *****
* Default font for this language. Should match font-family definitions
* in gamefont.css, or use specific font names; the former is more reliable.
* Can leave out if you just want this language to keep the base font.
*
* ***** MessageSpeed *****
* Message typing speed in this language. Default 1x (1 character per frame).
* Only specific speeds are supported; refer to the Message Speed (Base) setting
* or the associated comment (included in base export) for the possible options.
* Note that if you're using another plugin that alters message typing/speed,
* non-1x speeds may either override its functionality or not work properly.
*
*
* ----------------------------------------------------------------------
*  Defining Custom Values
* ----------------------------------------------------------------------
*
* Generated scripts come with a placeholder in the "Extra" section
* for user-defined variables. You can define custom values in Extra
* to help localize any kind of text not otherwise included.
*
* Define as many language-relevant values as you wish
* using the format "VariableName=value".
* (String values do not need to be contained in quotes.)
* These can be retrieved anywhere using the functions
* CTE.getString("VariableName") and CTE.getNumber("VariableName").
*
* As an example, say you want the font size in a plugin script
* to be different depending on the current language.
* While this could be accomplished without Extra definitions,
* it can often be more convenient to use them.
*
* Method 1: Directly checking the current language in the script.
* --------------------------------------------------
* [js/plugins/Plugin.js]
* if (ConfigManager.isLanguage("JP")) contents.fontSize = 24;
* else if (ConfigManager.isLanguage("EN")) contents.fontSize = 20;
* --------------------------------------------------
*
* Method 2: Defining an value in the Extra section and referring to it.
* --------------------------------------------------
* [languages/EN.txt]
* =====Extra=====
* PluginFontSize=20
* --------------------------------------------------
* [js/plugins/Plugin.js]
* contents.fontSize = CTE.getNumber("PluginFontSize", 24);
* --------------------------------------------------
*
* The second argument for getNumber in Method 2 is the default value,
* used for the Base Language, or if no value is found for the current language.
*
* CTE.getString() works the same way, but treats the value as a string.
* (Or more accurately, getNumber() automatically converts to an integer.)
*
* Example:
* --------------------------------------------------
* [languages/EN.txt]
* =====Extra=====
* GlossaryOptionName=Dictionary
* --------------------------------------------------
* [js/plugins/Plugin.js]
* contents.drawText(CTE.getString("GlossaryOptionName", "Glossary"), 0, 0);
* --------------------------------------------------
*
* This displays "Dictionary" in English and "Glossary" in the Base Language,
* or any other language that does not define GlossaryOptionName.
*
*
* ----------------------------------------------------------------------
*  Alternate Image Loading
* ----------------------------------------------------------------------
*
* For cases where images differ depending on the language
* (containing text that needs translation, generally),
* this plugin can automatically load the appropriate image
* for the current language.
*
* Method 1:
* Within the appropriate image folder (i.e. "pictures", "parallaxes"),
* create a subfolder for each non-base language code (i.e. "EN").
* Place translated images in those language subfolders,
* using the exact same filename as the base image,
* and they will be used automatically for that language.
*
* Method 2:
* If you name an image "name_XX", where XX is the Base Language Code,
* this plugin will automatically try to load the equivalent image
* for the current language code (i.e. "name_YY").
*
* Example:
* If there is a Show Picture command that displays the image "test_JP",
* when playing in English (language code EN), it will display "test_EN".
* (If no image named "test_EN" exists, it will fall back on "test_JP".)
*
* --------------------------------------------------
* WARNING
* Because only NW.js mode (standalone executable) has the ability to
* quickly check whether a file exists, this feature requires extra measures
* to function when the game is played in a browser.
*
* Make sure the "Make Image List" setting is enabled,
* then run the game in playtest mode (which uses NW.js).
* This auto-generates TranslatedImages.txt in the languages folder,
* a list of translated images that exist for each language.
* Besides generating on launch, it should also update
* if a new translated image gets loaded mid-game.
*
* As long as TranslatedImages.txt is up to date,
* image loading functionality should work the same in-browser.
* --------------------------------------------------
*
* Note that Alternate Image Loading applies to any kind of image,
* so it can even be used for things like tile or character graphics.
* However, not everything may update immediately on language change.
* (May be fixable with language refresh handler; see Customizable Handlers.)
*
*
* ----------------------------------------------------------------------
*  Language-Conditional Pages
* ----------------------------------------------------------------------
*
* If you need to do something complicated with events and language,
* like having a whole different event layout based on language,
* you can require a certain language for an event page to trigger.
*
* Add a comment to the top of the page that includes
* "<CTELanguage:EN>", and that page will only trigger/be visible in English.
* You can also use commas to specify multiple languages:
* "<CTELanguage:JP,EN>" will make it trigger in Japanese or English.
*
*
* ----------------------------------------------------------------------
*  Command Text Conditions
* ----------------------------------------------------------------------
*
* There may be times when you want to split a command
* into two or more cases based on some condition,
* but don't want to have to add a branch in the code itself,
* particularly since it may not be needed in every language.
* (A common instance of this is plurality for variable amounts.)
*
* You can use the syntax "＠if [condition]" and "＠else" inside a command
* to define conditions and the text to use if that condition is met.
*
* (Note: For technical reasons, this plugin description cannot use
* a normal "at sign" character. Thus, a "full-width" one is used here,
* but conditions allow either a normal or full-width "at sign" to be used.
* Sorry for the confusion.)
*
* Be aware that the FIRST satisfied condition in the list will be used,
* meaning higher-priority conditions should be given first.
* If no conditions at all are met, Base Language text will be used,
* so you should generally always define an "＠else" case.
*
* Example:
* #Message#
* ＠if V[1] = 1
* There is \V[1] item.
* ＠if V[1] >= 100
* There are a lot of items.
* ＠if V[1] >= 200
* There are too many items.
* ＠else
* There are \V[1] items.
* ##
*
* Variable 1 has value 1 -> "There is 1 item."
* Variable 1 has value 2 -> "There are 2 items."
* Variable 1 has value 99 -> "There are 99 items."
* Variable 1 has value from 100 to 199 -> "There are a lot of items."
* Variable 1 has value 200 or greater -> "There are a lot of items."
* ("V[1] >= 100" is reached and met first, so "V[1] >= 200" is never reached.)
*
* Accepted syntax for conditions:
* "S[##]" is converted to the value of Switch ##.
* "V[##]" is converted to the value of Variable ##.
* Numbers and string literals remain the same.
* "true"/"on" and "false"/"off" can be used for checking switches.
* Putting ! at the start of something will negate it if possible.
*
* Valid comparisons:
* Equal (=== or == or =)
* Not equal (!== or !=)
* Greater than or equal (>=)
* Less than or equal (<=)
* Greater than (>)
* Less than (<)
* If no comparison is stated, it checks if the value is true.
* (Example: "＠if S1" -> Switch 1 is on. "＠if !S1" -> Switch 1 is off.)
*
*
* ----------------------------------------------------------------------
*  Logging Identical Strings
* ----------------------------------------------------------------------
*
* The plugin can compare text in each language with its counterparts,
* and check if that text is identical in two or more languages,
* indicating that it may not have been translated.
* If anything is found, it is logged to "IdenticalWarnings.txt."
*
* This check will run on launch in playtest mode
* if you enable Launch Actions > Log Identical Strings,
* or you can choose to run it from the Export Actions menu
* (hold down the Script Reload key in-game in playtest mode).
*
* It can return many false positives in cases where
* there's good reason for text to just be the same,
* but it can help double-check that no text was overlooked.
* (The next section explains a way to cut down on false positives.)
*
*
* ----------------------------------------------------------------------
*  Ignoring Text Issues
* ----------------------------------------------------------------------
*
* If you know certain text is unused and need not be translated,
* you can specify areas to be "ignored" in a language script file.
* Ignored elements will be omitted from Identical String logs,
* and show as "IGNORED" instead of "NOT FOUND" in the Reference Table.
*
* (Naturally, if you want to be able to use identical string logging
* for its primary purpose of tracking down untranslated text,
* you should make very sure that you only ignore things
* that are definitely unused/unnecessary!)
*
* In most cases, you can simply add the line "%ignore"
* just after the marker for the section you want to ignore.
* The scope will be assumed based on file/event/page markers.
*
* Keep in mind that "ignoring" a section just means
* ignoring *warnings* if anything is identical or missing;
* it does not "invalidate" any text that you DO provide in that section.
* Thus, you're free to ignore a wide range and include only what is needed.
*
* Example: If only a handful of Terms are actually used by the game,
* you can add an "%ignore" line after the "=Terms=" header,
* and then include only the fields that need translating.
* You will not be bothered about the absence of other fields.
*
* Note, however, that missing translations encountered in-game
* will always be warned about (if those warnings are enabled).
* This is because the fact they can be encountered in-game
* indicates that they should have an explicit translation given.
*
* The following are some more specific forms of %ignore
* that may be convenient in certain cases.
*
* %ignorefile [file]
* Ignores issues within the given file.
* Valid files: Map##, CommonEvents, Troops, databases (Actors/Skills/etc.)
* Can also specify a range of maps with "MapXX-YY".
* If no file is given, the current file is assumed.
*
* %ignoreevent [event]
* Ignores issues within the given event in the current map.
* Can also specify a range of events with "XX-YY".
* If no event is given, the current event is assumed.
*
* %ignorepage [page]
* Ignores issues within the given page in the current event.
* Can also specify a range of pages with "XX-YY".
* If no page is given, the current page is assumed.
*
* %ignorecommand commandname [scope]
* Ignores issues for any instance of the given command name.
* Valid scopes: File, Event, Page
* If no scope is given, it is assumed based on file/event/page markers.
*
* %ignoreentry [entry]
* Ignores issues within given entry in the current database.
* Can also specify a range of entries with "XX-YY".
* If no entry is given, the current entry is assumed.
*
* %ignorefield fieldname
* Ignores issues for any instance of the given database field.
* This applies to all entries in the current database.
*
* %ignorethis
* Put on same line as a command/field header (i.e. "#Message# %ignorethis")
* to ignore warnings for that command/field only.
*
*
* ----------------------------------------------------------------------
*  Message Debugger Menu
* ----------------------------------------------------------------------
*
* In playtest mode, you can press the Message Debugger Key
* (if enabled; default F11) to open a message preview menu.
* This presents a list of all maps, common events, and troops
* which contain Message commands in the current language.
* You can then go in and select a message to preview.
*
* If you press the Message Debugger Key while the menu is open,
* it opens a search box to search the current list for a string.
*
* The language file is reloaded regularly while navigating the menu,
* allowing you to edit and quickly check messages.
*
* Message display settings for this menu can be customized
* by editing functions (see Customizable Handlers).
*
*
* ----------------------------------------------------------------------
*  Message Typer Window
* ----------------------------------------------------------------------
*
* Besides text in the language scripts, this plugin also provides
* a way to preview arbitrary messages in-game.
*
* From the Export Actions menu (hold down the Script Reload key
* in-game in playtest mode), you can open the Message Typer sub-window,
* which has a textbox you can type anything into and then preview
* using an in-game message window.
*
* Press Shift+Enter to enter line breaks.
* By default, text will type instantly instead of typing out,
* and if you edit the textbox while a preview message window is up,
* it will automatically update to show the new text.
* You can change these settings by clicking the radio buttons.
*
*
* ----------------------------------------------------------------------
*  Reformatting Language Files
* ----------------------------------------------------------------------
*
* This plugin offers a function to "reconstruct" language scripts;
* that is, re-export them in the exact format of a Base Language script,
* but keeping the actual text content from the given language.
*
* This can generally help you keep your language scripts "clean,"
* updating old comment info and removing errant linebreaks or symbols.
* It can be particularly useful if you make changes to context-aiding
* script comments added by the commandExportHandler functions.
* Additionally, by changing your Base Text Export settings,
* reformatting can split up all-in-one language files, or vice versa.
*
* From the Export Actions menu (hold down the Script Reload key
* in-game in playtest mode), you can re-export the current language's data
* in the same format as an up-to-date base text export,
* but using the latest text data for that language.
* It will be saved using the file/folder name "XX (Reformatted)"
* so as not to accidentally overwrite the original language files.
*
* Similar to the Reference Table, non-ignored text that is
* not found in the language data shows as "<<<NOT FOUND>>>".
* Meanwhile, excess commands get a comment of "<<<NOT FOUND IN BASE>>>",
* and their headers are converted from "#Message#" to "_Message_"
* to make it clear that text is not currently getting used in-game.
*
* Any comments not included as part of base text export will be lost,
* and the Variables and Extra sections in particular may
* have their order altered or otherwise become less readable.
* You may want to manually copy these over to the reformatted files.
*
*
* ----------------------------------------------------------------------
*  Customizable Handlers
* ----------------------------------------------------------------------
*
* The plugin code includes a few empty handler functions,
* intended to be customized to suit the needs of a specific game.
* You can either directly edit the content of the functions,
* or override them in your own plugin (loaded after the CTE plugin).
*
* ***** CTE.commandExportHandlerPrefix(command) *****
* ***** CTE.commandExportHandlerPostfix(command) *****
* To aid in translating, you may want more contextual information in scripts.
* These help you expand the script export function relatively easily.
*
* These functions get called for every command in a page,
* and are passed the command object itself as an argument.
* Based on "command.code" and "command.parameters",
* you can add export text for more commands.
* (Prefix goes before existing command text, Postfix comes after.)
*
* Example:
* Suppose you want to include information on the face graphics
* used in message boxes, to make it clear who's talking.
* This would be part of Command 101 (Show Text),
* which the base plugin already handles the message part of,
* but you can append this additional information as a prefix.
*
* CTE.commandExportHandlerPrefix = function(command) {
*    if (command.code == 101) { // Show Text
*       switch (command.parameters[0]) { // Face graphic filename
*       case "Actor1":
*          switch (command.parameters[1]) { // Face index
*             case 0: return "{Harold}\n\n";
*             case 1: return "{Therese}\n\n";
*             case 2: return "{Marsha}\n\n";
*             case 3: return "{Lucius}\n\n";
*             case 4: return "{...}\n\n;
*             case 5: return "{...}\n\n;
*             case 6: return "{...}\n\n;
*             case 7: return "{...}\n\n;
*          }
*          break;
*       case "Actor2":
*          switch (command.parameters[1]) { // Face index
*             // etcetera
*          }
*          break;
*       }
*    }
*    return ""; // Nothing to write
* };
*
* ***** CTE.languageRefreshHandler() *****
* This handler is called whenever the language changes.
* If something using language-based data does not update automatically,
* you can have this function call whatever necessary to make it refresh.
*
* ***** CTE.messageDebuggerPreviewTextHandler(message) *****
* This handler is called when listing messages in the Message Debugger.
*
* Messages are processed into "preview text" for display in the list,
* which includes removal of linebreaks and command codes like \C[X].
* If you want to remove additional command codes added by plugins,
* you can do so by processing the message text here before returning it.
* (The commented code gives an example for removing "\yourcodehere[___]".)
*
* ***** CTE.messageDebuggerMessageSettingsHandler(message, source) *****
* ***** CTE.messageDebuggerMessageSettingsPostHandler() *****
* The first is called before previewing a message with the Message Debugger,
* and the second is called after closing a previewed message.
*
* The first handler can set up various message settings before display,
* such as window type or position, or show a textbox background picture.
* (The default function shows an example for setting background type.)
* The second handler is generally for post-message cleanup,
* such as erasing pictures shown by the first handler.
*
* The first handler is passed the message content and the source,
* since these may be relevant for determining display settings.
* (Source is given as "map##", "commonevent##", or "troop##",
* where ## is the map number, common event number, or troop number.)
*
* ***** CTE.languageCodeFromExternalAPI() *****
* Handler to integrate external language setting via an API such as Steam.
*
* If this function returns a valid language code, it takes priority
* over Default Language as the language the game first launches in.
* Be sure to edit the language codes being returned
* if they don't match the language codes you're using.
*
* The default function should work for a few Steam API plugins,
* but it is theoretically modifiable for other launchers/APIs
* that have external language settings.
*
*
* ----------------------------------------------------------------------
*  Miscellaneous Information
* ----------------------------------------------------------------------
*
* If you want to change the current language in some way other than
* letting the user change it in the Options menu - for instance, making
* a custom language select screen - you can simply set the language
* in a Script command or plugin file like so:
* ConfigManager.language = "EN";
*
* All update processes associated with changing language are triggered
* even when setting ConfigManager.language directly like this,
* so everything should behave generally the same as it does when
* the user changes the language in the Options menu.
*
*
* ======================================================================
*                                Credits
* ======================================================================
*
* Created by vgperson.
*
* Special thanks to SumRndmDde's Translation Engine plugin,
* which was partially referred to for the creation of this plugin.
*
* Current Version: 1.06
*
* --- Changelog ---
*
* - Version 1.06 -
* Fixed failure to save game in RPG Maker MZ games.
*
* - Version 1.05 -
* Made some changes to CTE.load and CTE.loadTranslatedImagesFile
* which should allow for loading files in mobile ports.
* Fixed a bug where individual Troop scripts would not be loaded properly.
*
* - Version 1.04a -
* Added Speakers section to globally translate MZ speaker names.
* Speaker names are also included as comments above messages on export.
*
* - Version 1.04 -
* Made some fixes for (partial) RPG Maker MZ support.
* The MV-only font loader function won't be called if it doesn't exist,
* and command text replacement no longer uses the _params variable
* which is absent in MZ.
*
* - Version 1.03 -
* Added support for GALV_TimedMessagePopups plugin.
* Previously, CTE was overriding that plugin such that localized messages
* tagged as popups would just display as regular messages.
*
* - Version 1.02a -
* Fixed an oversight that caused individual System and Classes
* language scripts to not be loaded.
*
* - Version 1.02 -
* Reworked choice text localization, adding built-in support for two plugins
* that combine consecutive choice commands into one big list.
* It should also be relatively easy to add support for other plugins
* that affect choices, by calling the localizeChoiceCommand function
* before using that command's parameters.
*
* - Version 1.01 -
* Changed Base Text Export to leave Language Display Name core setting blank.
* In addiiton, if Language Display Name is blank, it defaults to language code.
*
* Since it's common to rename base language scripts to start a new language,
* this helps avoid the confusing situation of language names showing as
* the default language until manually changed, making it seem as if
* the option is not actually changing.
*
* - Version 1.00 -
* First official release.
*/

/******************************
USER SETTINGS
******************************/

var CTE = CTE || {};

CTE.tryParseList = function(json, defaultResult = []) {
	try {
		var result = JSON.parse(json);
		if (result.length == 0) return defaultResult;
		return result;
	}
	catch (e) {
		return defaultResult;
	}
};

CTE.tryParseBoolean = function(json, defaultResult = false) {
	try {
		return JSON.parse(json.toLowerCase());
	}
	catch (e) {
		return defaultResult;
	}
};

// Return whether game is in playtest mode, overriding with force-playtest setting.
CTE.isPlaytest = function() {
	return CTE.ForcePlaytestMode || (Utils.isNwjs() && Utils.isOptionValid("test"));
}

var pluginParams = PluginManager.parameters("CustomTranslationEngine");

CTE.DefaultLanguageCodes = CTE.tryParseList(pluginParams["DefaultLanguageCodes"], ["EN"]);
CTE.BaseLanguage = pluginParams["BaseLanguage"];
CTE.DefaultLanguage = pluginParams["DefaultLanguage"];

CTE.BaseLanguageDisplayName = pluginParams["BaseLanguageDisplayName"] || "English";
CTE.BaseLanguageOptionName = pluginParams["BaseLanguageOptionName"] || "Language";
CTE.PluginDatabases = CTE.tryParseList(pluginParams["PluginDatabases"], []);
CTE.BaseMessageSpeed = pluginParams["BaseMessageSpeed"] || "1x";

CTE.ForcePlaytestMode = CTE.tryParseBoolean(pluginParams["ForcePlaytestMode"]);

if (pluginParams["LoadCTEFiles"] === "txtplaytest") { CTE.LoadCTEFiles = !CTE.isPlaytest(); }
else { CTE.LoadCTEFiles = CTE.tryParseBoolean(pluginParams["LoadCTEFiles"]); }
CTE.GenerateCTEFiles = CTE.tryParseBoolean(pluginParams["GenerateCTEFiles"]);

CTE.ExportBaseOnLaunch = CTE.tryParseBoolean(pluginParams["ExportBaseOnLaunch"]);
CTE.BaseExportFilename = pluginParams["BaseExportFilename"];

CTE.AllInOneFile = CTE.tryParseBoolean(pluginParams["AllInOneFile"], true);
CTE.AllInOneMaps = CTE.tryParseBoolean(pluginParams["AllInOneMaps"]);
CTE.AllInOneCommons = CTE.tryParseBoolean(pluginParams["AllInOneCommons"], true);
CTE.AllInOneTroops = CTE.tryParseBoolean(pluginParams["AllInOneTroops"], true);
CTE.AllInOneDatabase = CTE.tryParseBoolean(pluginParams["AllInOneDatabase"], true);
CTE.NotetagsToExport = CTE.tryParseList(pluginParams["NotetagsToExport"], []);
CTE.PluginParamsToExport = CTE.tryParseList(pluginParams["PluginParamsToExport"], []);

CTE.ExportTableOnLaunch = CTE.tryParseBoolean(pluginParams["ExportTableOnLaunch"]);
CTE.ExportIdenticalOnLaunch = CTE.tryParseBoolean(pluginParams["ExportIdenticalOnLaunch"]);

CTE.MakeImageList = CTE.tryParseBoolean(pluginParams["MakeImageList"], true);
if (pluginParams["LoadAnyLanguageFiles"] === "notinplaytest") { CTE.LoadAnyLanguageFiles = !CTE.isPlaytest(); }
else { CTE.LoadAnyLanguageFiles = CTE.tryParseBoolean(pluginParams["LoadAnyLanguageFiles"]); }
CTE.LockLanguage = CTE.tryParseBoolean(pluginParams["LockLanguage"]);
CTE.ReloadKey = pluginParams["ReloadKey"];
CTE.MessageDebugKey = pluginParams["MessageDebugKey"];
CTE.LogSetting = pluginParams["LogSetting"];

if (CTE.ReloadKey < 0) {
	CTE.ReloadKey = 0;
}
if (CTE.MessageDebugKey < 0) {
	CTE.MessageDebugKey = 0;
}

CTE.languagesLoaded = false;
CTE.launchLoadPending = true;
CTE.launchExportPending = true;
CTE.launchImageListPending = false;

CTE.localizeVariableOverride = false;

CTE.preloadedFonts = [];
CTE.translatedImageList = {};

$CTETemp = {};

/******************************
FILE LOADING FUNCTIONS
******************************/

// Loads a file into CTE.languageObject, returning true if file exists and load was successful.
CTE.load = function(localPath) {
	CTE.languageObject = {};
	CTE.lastLoadedFile = "";
	CTE.loadStatus = 0;

	try {
		const xhr = new XMLHttpRequest();
		xhr.open("GET", localPath, false);
		xhr.overrideMimeType("application/json");
		xhr.onload = () => CTE.onLanguageLoad(xhr, localPath);
		xhr.onerror = () => {
			CTE.warn(xhr.status + " error loading " + localPath + ".");
			CTE.loadStatus = -1;
		};
		xhr.send();
	}
	catch (e) {
		CTE.loadStatus = -1;
	}

	return CTE.loadStatus == 1;
};

CTE.onLanguageLoad = function(xhr, url) {
	if (xhr.status < 400) {
		CTE.languageObject = JSON.parse(xhr.responseText);
		CTE.languageObjectUrl = url;
		CTE.loadStatus = 1;
	}
	else {
		CTE.warn(xhr.status + " error loading " + url + ".");
		CTE.loadStatus = -1;
	}
}

/******************************
DATA SETTERS/GETTERS
******************************/

// Set language data for given language.
CTE.setDataForLanguage = function(language, value, ...path) {
	CTE.data = CTE.data || {};
	CTE.data[language] = CTE.data[language] || {};
	const lastKey = path.pop();
	const data = path.reduce((data, segment) => data[segment] = data[segment] || {}, CTE.data[language]);
	if (lastKey == null) {
		Object.assign(data, value)
	} else {
		data[lastKey] = value;
	}
};

// Get language data for current language.
CTE.getData = function(...path) {
	return CTE.getDataForLanguage(ConfigManager.getLanguage(), ...path);
};

// Get language data for given language.
CTE.getDataForLanguage = function(language, ...path) {
	const data = path.reduce((data, segment) => data?.[segment], CTE.data?.[language]);

	if (Array.isArray(data)) {
		// Data is conditional text array; check conditions to determine which text to use
		const conditionalData = data.find(data => CTE.conditionMet(data[0]))
		return conditionalData ? conditionalData[1] : null;
	} else {
		return data;
	}
};

// Return whether the language data exists for current language.
CTE.dataExists = function(...path) {
	return CTE.dataExistsForLanguage(ConfigManager.getLanguage(), ...path);
};

// Return whether the language data exists for given language.
CTE.dataExistsForLanguage = function(language, ...path) {
	return !!CTE.getDataForLanguage(language, ...path);
};

// Return whether there is a regex list for the target variable in the current language.
CTE.variableRegexExists = function(targetVariable) {
	return CTE.dataExists("variables", targetVariable, "__regex__");
};

// Process regex list for the target variable in the current language, returning result (the original string, if nothing matches).
CTE.processVariableRegex = function(targetVariable, str) {
	var regexList = CTE.getData("variables", targetVariable, "__regex__");
	for (var i = 0; i < regexList.length; i++) {
		var regex = new RegExp("^" + regexList[i][0] + "$");
		if (str.match(regex)) {
			var translation = regexList[i][1];
			var args = str.match(regex);
			if (args.length > 1) {
				for (var k = 1; k < args.length; k++) {
					translation = translation.replace(new RegExp("\\{" + (k - 1) + "\\}", "g"), args[k]);
				}
			}
			return translation;
		}
	}
	return str;
};

// Get string data defined in the Extra section for current/specified language, using default value if not found.
CTE.getString = function(name, defaultValue = "", language = null) {
	if (ConfigManager.isBaseLanguage() && language !== null) return defaultValue;
	if (language === null) language = ConfigManager.getLanguage();

	if (CTE.dataExistsForLanguage(language, "extra", name)) {
		return CTE.getDataForLanguage(language, "extra", name).replace(/\\n/g, "\n");
	}
	return defaultValue;
};

// Get number data defined in the Extra section for current/specified language, using default value if not found.
CTE.getNumber = function(name, defaultValue = 0, language = null) {
	if (ConfigManager.isBaseLanguage() && language !== null) return defaultValue;
	const value = CTE.getString(name, defaultValue, language);
	return parseInt(value, 10);
};

// Set section to ignore warnings in for given language.
CTE.setIgnore = function(language, category, fieldName = "") {
	CTE.data = CTE.data || {};
	CTE.data[language] = CTE.data[language] || {};
	CTE.data[language]["ignore"] = CTE.data[language]["ignore"] || [];
	CTE.data[language]["ignore"].push(category + (fieldName !== "" ? "_" + fieldName : ""));
}

// Return whether to ignore warnings for text in the given location.
// category of "f-e-p" or shorter will auto-convert to exporting position.
// thisLevelOnly checks ignore definition for given level only, but otherwise, it checks all higher-level definitions as well.
CTE.shouldIgnore = function(language, category, field, thisLevelOnly = false) {
	if (category === "") return false;

	if (typeof CTE.data === "undefined"
		|| typeof CTE.data[language] === "undefined"
		|| typeof CTE.data[language]["ignore"] === "undefined")
		return false;

	if (typeof category === "string") category = category.toLowerCase();
	if (typeof field === "string") field = field.toLowerCase();

	if (category.startsWith("f")) { // Shorthand for current exporting position
		var position = CTE.exportingPosition;
		switch (category) {
			case "f": category = position["file"]; break;
			case "f-e": category = position["file"] + "-" + position["event"]; break;
			case "f-e-p": category = position["file"] + "-" + position["event"] + "-" + position["page"]; break;
		}
	}

	var list = CTE.data[language]["ignore"];
	var loop = true;
	while (loop) {
		if (field !== "" && list.includes(category + "_" + field)) return true;
		if (list.includes(category)) return true;

		if (!thisLevelOnly && category.includes("-")) { // Remove one level from end of category to check more general category next loop
			category = category.substring(0, category.lastIndexOf("-"));
		}
		else {
			loop = false;
		}
	}

	return false;
};

/******************************
LOADING LANGUAGE FILES
******************************/

// Load language scripts for all codes except base language.
CTE.loadLanguageScripts = function() {
	// Load the language codes
	const languagesPath = CTE_getLanguagesFolder();
	CTE.LanguageCodes = CTE.load(languagesPath + "Languages.json")
		? CTE.languageObject
		: CTE.DefaultLanguageCodes;

	// Only allow for image list generation after language codes have been loaded
	CTE.launchImageListPending = true;

	// Try to make unusual settings something valid.
	if (!CTE.LanguageCodes.includes(CTE.BaseLanguage)) {
		const baseFound = CTE.LanguageCodes.some(code => code.toLowerCase() === CTE.BaseLanguage.toLowerCase()); // Check case-insensitively as well
		if (!baseFound) CTE.BaseLanguage = CTE.LanguageCodes[0];
	}

	if (!CTE.LanguageCodes.includes(CTE.BaseLanguage)) {
		const defaultFound = CTE.LanguageCodes.some(code => code.toLowerCase() === CTE.DefaultLanguage.toLowerCase()); // Check case-insensitively as well
		if (!defaultFound) CTE.DefaultLanguage = CTE.BaseLanguage;
	}

	// Load known languages.
	CTE.LanguageCodes.filter(code => code.toLowerCase() !== CTE.BaseLanguage.toLowerCase())
		.forEach(code => CTE.loadLanguage(code));

	// Check for possible additional language files to load, and add to language code list if valid.
	if (CTE.LoadAnyLanguageFiles && Utils.isNwjs()) {
		const fs = require("fs");
		const path = require("path");
		const fullPath = path.join(path.dirname(process.mainModule.filename), languagesPath);

		const files = fs.readdirSync(fullPath);
		const processedFiles = [CTE.BaseLanguage.toLowerCase(), "translatedimages", "identicalwarnings"];
		for (let fileIndex = 0; fileIndex < files.length; fileIndex++) {
			const file = files[fileIndex];
			const isDirectory = fs.lstatSync(path.join(fullPath, file)).isDirectory();

			if (isDirectory || file.toLowerCase().endsWith(".cte") || file.toLowerCase().endsWith(".json")) {
				const possibleCode = !isDirectory ? file.slice(0, -4) : file;

				if (!CTE.LanguageCodes.contains(possibleCode)
					&& !processedFiles.includes(possibleCode.toLowerCase())) {
					CTE.addedLanguageValid = false;
					CTE.loadLanguage(possibleCode, false, true);

					if (CTE.addedLanguageValid) { // Add to language code list
						CTE.LanguageCodes.push(possibleCode);
						CTE.warn("Loaded " + possibleCode + " as an additional language.");
					} else if (CTE.data[possibleCode]) { // Remove any data that was made (usually just an empty object)
						delete CTE.data[possibleCode];
					}
				}
			}
		}
	}

	CTE.languagesLoaded = true;
	CTE.languageRefreshHandler();
};

// Load data for a language into CTE.data[language]. cteFallback is used if no TXTs are found in TXT mode.
// checkingAddedLanguage is used when attempting to load a potential additional language that may or may not be valid.
CTE.loadLanguage = function(language, cteFallback = false, checkingAddedLanguage = false) {
	const languagesPath = CTE_getLanguagesFolder();
	CTE.data = CTE.data || {};
	CTE.data[language] = {};
	CTE.anyFileModified = false;
	CTE.anyFileLoaded = false;

	if (CTE.LoadCTEFiles || cteFallback) {
		try {
			if (CTE.load(languagesPath + language + ".cte")) {
				CTE.data[language] = JSON.parse(CTE.languageObject);
				if (cteFallback && !checkingAddedLanguage) CTE.warn("Loaded " + language + ".cte as fallback.");
				if (checkingAddedLanguage) CTE.addedLanguageValid = true;
				CTE.preloadDefaultFont(language);
			} else if (!checkingAddedLanguage) {
				CTE.warn(language + ".cte could not be found.");
			}
		} catch (e) {
			if (!checkingAddedLanguage) {
				CTE.warn("Failed to load " + language + ".cte. " + e.message);
			}
		}
		return;
	}

	const languageFolder = languagesPath + language + "/";
	const loadAndProcess = (file, type, key) => {
		if (CTE.load(languageFolder + file)) CTE.processLanguageFile(language, type, key);
	};

	if (CTE.AllInOneFile && CTE.load(languagesPath + language + ".json")) {
		CTE.processLanguageFile(language);
	} else {
		if (CTE.AllInOneMaps) loadAndProcess("AllMaps.json", "map");
		else $dataMapInfos.forEach(map => map && loadAndProcess("Map" + map.id.padZero(3) + ".json", "map", "map" + map.id));

		if (CTE.AllInOneCommons) loadAndProcess("AllCommons.json", "common", "commonevents");
		else $dataCommonEvents.forEach(event => event && loadAndProcess("Common" + event.id.padZero(4) + ".txt", "common", "commonevents", event.id));

		if (CTE.AllInOneTroops) loadAndProcess("AllTroops.json", "troop", "troops");
		else $dataTroops.forEach(troop => troop && loadAndProcess("Troop" + troop.id.padZero(4) + ".json", "troop", "troops", troop.id));

		if (CTE.AllInOneDatabase) loadAndProcess("AllDatabases.json", "database");
		else Object.keys(CTE.getDatabases()).forEach(db => {
			if (!CTE["AllInOne" + db]) loadAndProcess(db + ".json", "database", db.toLowerCase());
		});

		loadAndProcess("AllPlugins.json", "plugins");
	}

	if (!CTE.anyFileLoaded) {
		if (!checkingAddedLanguage) CTE.warn("No language files found for " + language + ". Checking for .cte fallback...");
		CTE.loadLanguage(language, true, checkingAddedLanguage);
		return;
	}

	if (checkingAddedLanguage) CTE.addedLanguageValid = true;
	CTE.preloadDefaultFont(language);

	if (Utils.isNwjs() && CTE.isPlaytest() && CTE.GenerateCTEFiles) {
		// In playtesting, export loaded language data as CTE file (encrypted JSON) if enabled
		if (!CTE.anyFileModified) {
			// No files modified since last load, note that data has not updated since last time
			if (Object.keys(CTE.data[language]).length > 0) CTE.warn(language + ".cte is already up to date.");
		} else {
			try {
				const fs = require("fs");
				const path = require("path");
				const fullPath = path.join(path.dirname(process.mainModule.filename), languagesPath);
				const json = JSON.stringify(CTE.data[language], null, 4);
				const data = LZString.compressToBase64(json);
				fs.writeFileSync(fullPath + language + ".cte", data);
				CTE.warn("Exported " + language + ".cte using latest data.");
			} catch (e) {
				CTE.warn("Failed to write " + language + ".cte. " + e.message);
			}
		}
		if (CTE.GenerateCTEFiles_backup) {
			CTE.GenerateCTEFiles = CTE.GenerateCTEFiles_backup;
			CTE.GenerateCTEFiles_backup = undefined;
		}
		if (CTE.LoadCTEFiles_backup) {
			CTE.LoadCTEFiles = CTE.LoadCTEFiles_backup;
			CTE.LoadCTEFiles_backup = undefined;
		}
	}
};

CTE.conditionRegexes = [
	{ regex: /v\[(\d+)]/ig, replacement: `$gameVariables?.value(%1)` },
	{ regex: /s\[(\d+)]/ig, replacement: `$gameSwitches?.value(%1)` },
]

// Function to parse conditional text
CTE.parseConditionalText = function(text) {
	const parsedVariants = [];
	text = text.replace(/＠/g, "@"); // Normalize full-width '@'

	if (text.toLowerCase().includes("@if ") || text.toLowerCase().includes("@else")) {
		const lines = text.split("\n");
		let currentCondition = "";
		let subContent = null;

		for (let line of lines) {
			const trimmedLine = line.toLowerCase().trim();

			if (trimmedLine.startsWith("@if ")) {
				if (subContent) {
					const condition = new Function("context", `return ${currentCondition || true}`);
					parsedVariants.push([condition, subContent]);
				}
				currentCondition = line.substring(4).trim(); // Store condition
				subContent = null;
			} else if (trimmedLine.startsWith("@else")) {
				if (subContent) {
					const condition = new Function("context", `return ${currentCondition || true}`);
					parsedVariants.push([condition, subContent]);
				}
				currentCondition = "";
				subContent = null;
			} else {
				subContent = !subContent ? line : subContent + "\n" + line;
			}

			if (currentCondition) {
				for (const conditionRegex of CTE.conditionRegexes) {
					currentCondition = currentCondition.replace(conditionRegex.regex, (_, captured) => {
						// First parameter _ is full match which we don't need
						return conditionRegex.replacement.format(captured);
					});
				}
			}
		}

		if (subContent) {
			const condition = new Function("context", `return ${currentCondition || true}`);
			parsedVariants.push([condition, subContent]);
		}
	}

	return parsedVariants.length > 0 ? parsedVariants : text;
};

// Traverse the leaves of a JSON object. Returns true if the traversal should be continued
CTE.traverseLeaves = (node, key, parent, leafHandler) => {
	if (typeof node === "string") {
		return leafHandler(node, key, parent);
	} else if (node !== null) {
		for (const k in node) {
			if (!CTE.traverseLeaves(node[k], k, node, leafHandler)) {
				return false;
			}
		}
	}

	return true;
};

// After loading an individual language file into CTE.languageObject, process it and add to language data.
CTE.processLanguageFile = function(language, initialMode = "", initialFile = "", initialEvent = 1) {
	if (Utils.isNwjs() && CTE.isPlaytest() && CTE.lastLoadedFile !== "") {
		const fs = require("fs");

		CTE.fileModifiedTime = CTE.fileModifiedTime || {};
		const modified = fs.statSync(CTE.lastLoadedFile).mtime.getTime();
		if (CTE.fileModifiedTime[CTE.lastLoadedFile] === undefined || CTE.fileModifiedTime[CTE.lastLoadedFile] !== modified) { // First load, or file modified since last load
			CTE.fileModifiedTime[CTE.lastLoadedFile] = modified;
			CTE.anyFileModified = true;
		}
	}

	const json = CTE.languageObject;
	if (Object.keys(json).length >= 0) {
		CTE.setDataForLanguage(language, json, initialFile != "" ? initialFile : undefined);
		CTE.anyFileLoaded = true;

		// Parse conditional text
		CTE.traverseLeaves(json, null, null, (node, key, parent) => {
			parent[key] = CTE.parseConditionalText(node);
			return true;
		});
	}
};

// In playtest mode, open message debugger when debugger key is pressed, reload scripts when reload key is pressed, and show export options when reload key is held.
CTE.Scene_Map_update = Scene_Map.prototype.update;
Scene_Map.prototype.update = function() {
	CTE.Scene_Map_update.apply(this, arguments);

	if (CTE.isPlaytest()) {
		if (CTE.ReloadKey !== 0 && !Input.keyMapper[CTE.ReloadKey]) Input.keyMapper[CTE.ReloadKey] = "languagereload";
		if (CTE.MessageDebugKey !== 0 && !Input.keyMapper[CTE.MessageDebugKey]) Input.keyMapper[CTE.MessageDebugKey] = "messagedebug";

		if (Input.isTriggered("messagedebug") && !!SceneManager._scene._mapLoaded && !$gameMap.isEventRunning() && !CTE.TextDebug.open) { // Open message debugger
			SoundManager.playOk();
			CTE.loadLanguageScripts();
			CTE.TextDebug.open = true;
			CTE.TextDebug.currentFile = "";
			SceneManager.push(Scene_TextDebugger);
		}
		else if (Input.isTriggered("languagereload")) {
			SoundManager.playOk();
			if (CTE.LoadCTEFiles) { // For this case of reloading for debug, it's most convenient to just load from .txts and renew .ctes
				CTE.LoadCTEFiles_backup = CTE.LoadCTEFiles;
				CTE.LoadCTEFiles = false;
				CTE.GenerateCTEFiles_backup = CTE.GenerateCTEFiles;
				CTE.GenerateCTEFiles = true;
			}
			CTE.warn("Reloading all .txt language scripts.");
			CTE.loadLanguageScripts();

			if ($gameMap && $gameMessage
				&& $gameMessage.hasText() && !$gameMessage.isChoice() && !$gameMessage.isNumberInput() && !$gameMessage.isItemChoice()
				&& !CTE.TextDebug.open && !CTE.MessageTyper.previewing) { // Actively displaying message; show it again with updated language data
				if (typeof SceneManager._scene._messageWindow !== "undefined") { // Reference does not exist in MZ
					SceneManager._scene._messageWindow.terminateMessage(); // Close message window
					CTE.currentMessageInterpreter._index = CTE.currentMessageIndex; // Set index back to start of message command
					CTE.currentMessageInterpreter.command101(); // Re-process message command
					SceneManager._scene._messageWindow.pause = false; // Clear waiting for input
					SceneManager._scene._messageWindow._waitCount = 0; // Clear pauses
					SceneManager._scene._messageWindow.startMessage(); // Show message
				}
			}
		}
		else if (Input.isLongPressed("languagereload") && $gameMap && $gameMessage && !$gameMessage.isBusy()) {
			$gameMap._interpreter._list = [];
			$gameMap._interpreter.setWaitMode("message");
			var choices = ["Export base language text", "Export IdenticalWarnings.txt", "Export .cte files for all languages",
				"Export reformatted language text", "Open message typer window", "Do nothing"];
			$gameMap._interpreter.setupChoices([choices, choices.length - 1, 0, 2, 0]);
			$gameMessage.setChoiceCallback(function(choice) {
				switch (choice) {
					case 0: // Export base language text
						CTE.exportingBase = true;
						CTE.beginExportProcess();
						break;
					case 1: // Export IdenticalWarnings.txt
						CTE.exportingIdentical = true;
						CTE.beginExportProcess();
						break;
					case 2: // Export latest .ctes
						CTE.GenerateCTEFiles_backup = CTE.GenerateCTEFiles;
						CTE.LoadCTEFiles_backup = CTE.LoadCTEFiles;
						CTE.LoadCTEFiles = false;
						CTE.GenerateCTEFiles = true;
						CTE.fileModifiedTime = {}; // Ensure all files are reloaded regardless of modified time
						CTE.loadLanguageScripts();
						break;
					case 3: // Export reformatted language text
						if (!ConfigManager.isBaseLanguage()) {
							CTE.languageForReformat = ConfigManager.getLanguage();
							CTE.loadLanguage(CTE.languageForReformat);
							CTE.exportingReformat = true;
							CTE.beginExportProcess();
						}
						else {
							CTE.warn("Reformatted exporting is unnecessary for Base Language.");
						}
						break;
					case 4: // Open message typer window
						CTE.openMessageTyperWindow();
						break;
				}
			});
		}
	}
};

/******************************
LANGUAGE TEXT OVERRIDES
******************************/

// Define ID of map being loaded before loading it so context can be set correctly.
CTE.DataManager_loadMapData = DataManager.loadMapData;
DataManager.loadMapData = function(mapId) {
	if (mapId > 0) CTE.loadingMapId = mapId;
	CTE.DataManager_loadMapData.apply(this, arguments);
};

// Called from DataManager.onLoad. After loading data files, either update text, or define context for command lists.
CTE.onDataLoad = function(object) {
	const databases = CTE.getDatabases();
	const dbName = Object.keys(databases).find(dbName => databases[dbName].updateDatabaseProperties && object === window["$data" + dbName]);
	if (dbName) CTE.updateDatabaseProperties(dbName, databases[dbName]);
	
	// Upon loading map, Common Events, or Troops, define contexts in the first command of every command list.
	if (object === $dataMap) {
		if ($dataMap && $dataMap.events) {
			for (let eventNum = 0; eventNum < $dataMap.events.length; eventNum++) {
				const event = $dataMap.events[eventNum];
				if (event !== null && event.pages) {
					for (let pageIndex = 0; pageIndex < event.pages.length; pageIndex++) {
						if (event.pages[pageIndex] !== null && event.pages[pageIndex].list) {
							const list = event.pages[pageIndex].list;
							list._context = "map";
							list._mapId = CTE.loadingMapId;
							list._eventId = event.id;
							list._pageIndex = pageIndex;
						}
					}
				}
			}
		}
	}

	if (object === $dataCommonEvents) {
		if ($dataCommonEvents) {
			for (let eventNum = 0; eventNum < $dataCommonEvents.length; eventNum++) {
				const event = $dataCommonEvents[eventNum];
				if (event !== null && event.list) {
					const list = event.list;
					list._context = "common";
					list._commonEventId = event.id;
				}
			}
		}
	}

	if (object === $dataTroops) {
		if ($dataTroops) {
			for (let troopNum = 0; troopNum < $dataTroops.length; troopNum++) {
				const troop = $dataTroops[troopNum];
				if (troop !== null && troop.pages) {
					for (let pageIndex = 0; pageIndex < troop.pages.length; pageIndex++) {
						if (troop.pages[pageIndex] !== null && troop.pages[pageIndex].list) {
							const list = troop.pages[pageIndex].list;
							list._context = "troop";
							list._troopId = troop.id;
							list._pageIndex = pageIndex;
						}
					}
				}
			}
		}
	}
};

CTE.ContextMap = {
	"common": (list) => ["commonevents", `common-${list._commonEventId}`],
	"troop": (list) => ["troops", `troop-${list._troopId}`, "pages", `${(list._pageIndex)}`],
	"map": (list) => [`map${list._mapId}`, "events", `${list._eventId}`, "pages", `${(list._pageIndex)}`]
}

// Get command position data for the given Game_Interpreter by extracting context variables from the first command in the list.
CTE.getPositionData = function(interpreter) {
	if (!interpreter || !interpreter._list) return [];
	const list = interpreter._list;
	const context = list._context;
	return CTE.ContextMap[context](list) || []
};

// Get command group identifier based on position and command type.
CTE.getCommandGroupId = function(position, commandName, commandNum) {
	position.push(commandName + "-" + commandNum);
};

// Get the instance number of the command at the given index in the given command list.
CTE.getCommandNum = function(list, index) {
	if (!list[index]) return -1;
	const myCode = list[index].code;
	let myCount = 0;
	for (let i = index - 1; i >= 0; i--) {
		if (list[i].code == myCode) {
			myCount++;
		}
	}
	return myCount;
};

// Evaluates a text condition.
CTE.conditionMet = function(condition) {
	if (!condition) return true;
	const context = CTE.createConditionContext();
	return condition.call(this, context);
};

CTE.createConditionContext = function() {
	const context = {};
	const eventId = $gameMap?.getCurrentInterpreter()?.eventId();
	if (eventId) {
		context.event = $gameMap.event(eventId)?.event();
	}
	
	const target = BattleManager?._action?._currentTarget;
	if (target) {
		context.target = target.isActor() ? target.actor() : target.enemy();
	}
	return context;
};

// Localize message text.
CTE.Game_Interpreter_command101 = Game_Interpreter.prototype.command101;
Game_Interpreter.prototype.command101 = function() {
	if (ConfigManager.isBaseLanguage()) { // In base language, use original function
		return CTE.Game_Interpreter_command101.apply(this, arguments);
	}

	if ($gameMessage.isBusy()) {
		return false;
	}

	const position = CTE.getPositionData(this);
	if (position[0] === "") return CTE.Game_Interpreter_command101.apply(this, arguments); // No context for command
	const commandNum = CTE.getCommandNum(this._list, this._index) + 1;
	CTE.getCommandGroupId(position, "message", commandNum);

	CTE.currentMessageInterpreter = this;
	CTE.currentMessageIndex = this._index;

	if (!CTE.dataExists(...position)) {
		CTE.warn("No translation data found for " + position + ".", 2);
		return CTE.Game_Interpreter_command101.apply(this, arguments);
	}

	const commandData = CTE.getData(...position);
	const baseSpeakerName = this._list[this._index].parameters[4];
	if (commandData === null) { // Data exists, but no valid condition met
		CTE.warn("No valid condition found for " + position + ".", 2);
	} else {
		while (this.nextEventCode() === 401) {
			// Skip through text data
			this._index++;
		}
	}

	const result = CTE.Game_Interpreter_command101.apply(this, arguments);

	if (commandData) {
		const speakerName = CTE.getData("speakers", baseSpeakerName) || baseSpeakerName;
		$gameMessage.setSpeakerName(speakerName);
		const lines = commandData.split("\n");
		lines.forEach(line => $gameMessage.add(line));
	}

	return result;
};

// Localize strings for the given choice command (or by default, the _index) in the given Game_Interpreter.
CTE.localizeChoiceCommand = function(interpreter, index = -1) {
	if (!ConfigManager.isBaseLanguage()) {
		if (index == -1) index = interpreter._index;
		const position = CTE.getPositionData(interpreter);
		if (position[0] === "") return; // No context for command
		const commandNum = CTE.getCommandNum(interpreter._list, index) + 1;
		CTE.getCommandGroupId(position, "choice", commandNum);

		if (CTE.dataExists(...position)) {
			const commandData = CTE.getData(...position);
			if (commandData === null) { // Data exists, but no valid condition met
				CTE.warn("No valid condition found for " + position + ".", 2);
			}
			else {
				const choices = commandData.split("\n");
				if (choices.length == interpreter._list[index].parameters[0].length) {
					interpreter._list[index].parameters[0] = choices;
				}
				else {
					CTE.warn("Choice counts do not match in " + position + ".", 2);
				}
			}
		}
		else {
			CTE.warn("No translation data found for " + position + ".", 2);
		}
	}
};

// Inject logic to to localize choice setup
CTE.Game_Interpreter_setupChoices = Game_Interpreter.prototype.setupChoices;
Game_Interpreter.prototype.setupChoices = function() {
	CTE.localizeChoiceCommand(this);
	CTE.Game_Interpreter_setupChoices.apply(this, arguments);
}

// Localize regular choice commands.
CTE.Game_Interpreter_command102 = Game_Interpreter.prototype.command102;
Game_Interpreter.prototype.command102 = function() {
	CTE.localizeChoiceCommand(this);
	return CTE.Game_Interpreter_command102.apply(this, arguments);
};

// Localize scrolling text.
CTE.Game_Interpreter_command105 = Game_Interpreter.prototype.command105;
Game_Interpreter.prototype.command105 = function() {
	if (ConfigManager.isBaseLanguage()) { // In base language, use original function
		return CTE.Game_Interpreter_command105.apply(this, arguments);
	}

	if ($gameMessage.isBusy()) {
		return false;
	}

	const position = CTE.getPositionData(this);
	if (position[0] === "") return CTE.Game_Interpreter_command105.apply(this, arguments);
	const commandNum = CTE.getCommandNum(this._list, this._index) + 1;
	CTE.getCommandGroupId(position, "scrollingmessage", commandNum);

	CTE.currentMessageInterpreter = this;

	if (!CTE.dataExists(...position)) {
		CTE.warn("No translation data found for " + position + ".", 2);
		return CTE.Game_Interpreter_command105.apply(this, arguments);
	}

	const commandData = CTE.getData(...position);
	if (commandData === null) { // Data exists, but no valid condition met
		CTE.warn("No valid condition found for " + position + ".", 2);
	} else {
		while (this.nextEventCode() === 405) {
			// Skip through text data
			this._index++;
		}
	}

	const result = CTE.Game_Interpreter_command105.apply(this, arguments);

	if (commandData) {
		const lines = commandData.split("\n");
		lines.forEach(line => $gameMessage.add(line));
	}

	return result;
};

// Localize map display names. (Pass false to force base data name.)
Game_Map.prototype.displayName = function(useTranslation = true) {
	if (useTranslation) {
		return TextManager.localize($dataMap.displayName, `map${this.mapId()}`, "displayName");
	}

	return $dataMap.displayName;
};

// Localize actor names using lookup.
Game_Actor.prototype.name = function() {
	if (this._name === $dataActors[this.actorId()].name) {
		return TextManager.localize(this._name, "actors", `actor-${this.actorId()}`, "name");
	} else {
		return TextManager.localize(this._name, "namechange", this._name);
	}
};

// Localize actor nicknames using lookup.
Game_Actor.prototype.nickname = function() {
	if (this._nickname === $dataActors[this.actorId()].nickname) {
		// Unchanged from database definition
		return TextManager.localize(this._nickname, "actors", `actor-${this.actorId()}`, "nickname");
	} else {
		return TextManager.localize(this._nickname, "nicknamechange", this._nickname);
	}
};

// Localize actor profiles using lookup.
Game_Actor.prototype.profile = function() {
	if (this._nickname === $dataActors[this.actorId()].profile) {
		// Unchanged from database definition
		return TextManager.localize(this._profile, "actors", `actor-${this.actorId()}`, "profile");
	} else {
		return TextManager.localize(this._profile, "profilechange", this._profile);
	}
};

// Update properties of various databases to inject localization where it can't be handled directly
CTE.updateDatabaseProperties = function(dbName, db) {
	const data = eval(`$data${dbName}`);
	if (data === null) return;

	for (let i = 0; i < data.length; i++) {
		for (const rule of db.rules) {
			const entry = data[i];
			if (entry === null) continue;

			const category = [dbName.toLowerCase(), `${db.prefix}-${i}`];
			const originalGetter = Object.getOwnPropertyDescriptor(entry, rule);
			Object.defineProperty(entry, rule, {
				get: function() {
					const original = originalGetter.get ? originalGetter.get.call(entry) : originalGetter.value;
					return TextManager.localize(original, ...category, rule);
				},
				configurable: true
			});
		}
	}
}

// Generic localize function
TextManager.localize = function(original, ...path) {
	if (!ConfigManager.isBaseLanguage()) {
		if (CTE.dataExists(...path)) {
			return CTE.getData(...path);
		}
	}

	return original || '';
}

// Localize enemy names.
Game_Enemy.prototype.originalName = function() {
	return TextManager.localize(this.enemy().name, "enemies", `enemy-${this.enemyId()}`, "name");
};

// Localize basic terms.
TextManager.basic = function(basicId) {
	return this.localize($dataSystem.terms.basic[basicId], "system", "terms", "basic", basicId);
};

// Localize param terms.
TextManager.param = function(paramId) {
	return this.localize($dataSystem.terms.params[paramId], "system", "terms", "params", paramId);
};

// Localize command terms.
TextManager.command = function(commandId) {
	return this.localize($dataSystem.terms.commands[commandId], "system", "terms", "commands", commandId);
};

// Localize message terms.
TextManager.message = function(messageId) {
	if (!ConfigManager.isBaseLanguage()) {
		const pluginSplit = messageId.split("-");
		if (CTE.dataExists("plugins", ...pluginSplit)) {
			return CTE.getData("plugins", ...pluginSplit);
		}
	}

	return this.localize($dataSystem.terms.messages[messageId], "system", "terms", "messages", messageId);
};

// Localize currency unit.
Object.defineProperty(TextManager, 'currencyUnit', {
	get: function() {
		return this.localize($dataSystem.currencyUnit, "system", "currencyUnit");
	},
	configurable: true
});

// Localize skill type names.
Window_SkillType.prototype.makeCommandList = function() {
	if (this._actor) {
		const skillTypes = this._actor.skillTypes();
		for (const stypeId of skillTypes) {
			const name = TextManager.localize($dataSystem.skillTypes[stypeId], "system", "skillTypes", stypeId);
			this.addCommand(name, "skill", true, stypeId);
		}
	}
};

// Localize skill type names.
Window_ActorCommand.prototype.addSkillCommands = function() {
	const skillTypes = this._actor.skillTypes();
	for (const stypeId of skillTypes) {
		const name = TextManager.localize($dataSystem.skillTypes[stypeId], "system", "skillTypes", stypeId);
		this.addCommand(name, "skill", true, stypeId);
	}
};

// Localize equip slot names.
Window_EquipSlot.prototype.slotName = function(index) {
	if (this._actor) {
		const slot = this._actor.equipSlots()[index];
		return TextManager.localize($dataSystem.equipTypes[slot], "system", "equipTypes", slot);
	}
	return '';
};

// Localize game title on boot.
Scene_Boot.prototype.updateDocumentTitle = function() {
	let gameTitle = CTE.getData("system", "gameTitle");
	if (gameTitle == null && $dataSystem) gameTitle = $dataSystem.gameTitle;
	if (gameTitle != null) document.title = gameTitle;
};

// Add a "localize" argument for basic variable getter, and an override to force localization of strings.
CTE.Game_Variables_value = Game_Variables.prototype.value;
Game_Variables.prototype.value = function(variableId, localize = false) {
	if ((localize || CTE.localizeVariableOverride) && this._data[variableId]) return CTE.localizeVariable(this._data[variableId], variableId);
	return CTE.Game_Variables_value.call(this, variableId);
};

// Localize string variable based on variable translation definitions, returning original value if nothing defined.
CTE.localizeVariable = function(str, targetVariable = 0) {
	if (typeof str !== "string") return str;

	for (const varId of [targetVariable, 0]) {
		if (CTE.dataExists("variables", varId, str)) return CTE.getData("variables", varId, str);
		const result = CTE.variableRegexExists(varId) ? CTE.processVariableRegex(varId, str) : str;
		if (result !== str) return result;
	}

	return str;
};

// Set variable localize override when converting variable codes in messages.
CTE.Window_Base_convertEscapeCharacters = Window_Base.prototype.convertEscapeCharacters;
Window_Base.prototype.convertEscapeCharacters = function(text) {
	CTE.localizeVariableOverride = true;
	const result = CTE.Window_Base_convertEscapeCharacters.apply(this, arguments);
	CTE.localizeVariableOverride = false;
	return result;
};

CTE.DataManager_savefileInfo = DataManager.savefileInfo;
DataManager.savefileInfo = function(savefileId) {
	const saveInfo = CTE.DataManager_savefileInfo.apply(this, arguments);
	if (!saveInfo) return saveInfo;

	const isBaseLang = ConfigManager.isBaseLanguage();

	for (const key of ["title", "mapName"]) {
		const backupKey = `_backup${key.charAt(0).toUpperCase()}${key.slice(1)}`;

		if (!isBaseLang) {
			if (saveInfo[backupKey] === undefined) saveInfo[backupKey] = saveInfo[key];

			if (key === "title" && CTE.dataExists("system", "title") && $dataSystem && saveInfo[backupKey] === $dataSystem.gameTitle) {
				saveInfo[key] = CTE.getData("system", "title");
			}
			if (key === "mapName" && CTE.dataExists("displayname", saveInfo[backupKey])) {
				saveInfo[key] = CTE.getData("displayname", saveInfo[backupKey]);
			}
		} else if (saveInfo[backupKey] !== undefined) {
			saveInfo[key] = saveInfo[backupKey];
		}
	}

	return saveInfo;
};

// When making save file info, keep non-translated map display name in a backup variable in case game uses a plugin to show display name.
CTE.DataManager_makeSavefileInfo = DataManager.makeSavefileInfo;
DataManager.makeSavefileInfo = function() {
	const info = CTE.DataManager_makeSavefileInfo.apply(this, arguments);
	info._backupMapName = $gameMap.displayName(false);
	return info;
}

// Just before saving to global file (so that ordering of makeSavefileInfo function chain doesn't matter), revert title and map mame to Base Language backups, then remove the backup variables.
CTE.DataManager_saveGlobalInfo = DataManager.saveGlobalInfo;
DataManager.saveGlobalInfo = function(info) {
	const properties = ["title", "mapName"];
	if (Array.isArray(info)) {
		for (let i = 0; i < info.length; i++) {
			if (!info[i]) continue;
			for (const property of properties) {
				const backupProperty = `_backup${key.charAt(0).toUpperCase()}${key.slice(1)}`;
				if (fileInfo[backupProperty]) {
					fileInfo[property] = fileInfo[backupProperty];
					delete fileInfo[backupProperty];
				}
			}
		}
	}
	CTE.DataManager_saveGlobalInfo.call(this, info);
};

/******************************
ALTERNATE IMAGE LOADING
******************************/

// Automatically try to load translated images for current language.
CTE.ImageManager_loadBitmap = ImageManager.loadBitmap;
ImageManager.loadBitmap = function(folder, filename, hue, smooth) {
	if (!ConfigManager.isBaseLanguage()) {
		// Determine if a translated version of the image exists in a subfolder for the current language.
		var language = ConfigManager.getLanguage();
		var imgSubfolder = folder.substr(folder.indexOf("img/") + "img/".length);
		var existsInList = false;
		if (typeof CTE.translatedImageList[language.toLowerCase()] !== "undefined" && CTE.translatedImageList[language.toLowerCase()].includes(imgSubfolder + filename)) {
			existsInList = true;
		}

		var encrypted = typeof Decrypter !== "undefined" ? !!Decrypter.hasEncryptedImages : false;

		if (Utils.isNwjs()) { // Directly check if file exists
			var fs = require("fs");
			var path = require("path");
			var imageInLanguageSubfolder = path.join(path.dirname(process.mainModule.filename), folder + language + "/" + filename + (!encrypted ? ".png" : ".rpgmvp"));
			if (fs.existsSync(imageInLanguageSubfolder)) {
				if (!existsInList && CTE.isPlaytest() && CTE.MakeImageList) { // New file discovered; add it to list and re-save
					CTE.translatedImageList[language.toLowerCase()] = CTE.translatedImageList[language.toLowerCase()] || [];
					CTE.translatedImageList[language.toLowerCase()].push(imgSubfolder + filename);
					CTE.saveTranslatedImagesFile();
				}
				try {
					return CTE.ImageManager_loadBitmap.call(this, folder + language + "/", filename, hue, smooth);
				}
				catch (e) {
				}
			}
		}
		else { // Refer to list of known translated images
			if (existsInList) {
				try {
					return CTE.ImageManager_loadBitmap.call(this, folder + language + "/", filename, hue, smooth);
				}
				catch (e) {
				}
			}
		}

		// Otherwise, if file ends in _XX (Base Language Code), check for equivalent _YY (current language code).
		if (filename.match(new RegExp("_" + CTE.BaseLanguage + "$"))) {
			var newName = filename.slice(0, -CTE.BaseLanguage.length) + ConfigManager.getLanguage();
			existsInList = false;
			if (typeof CTE.translatedImageList[language.toLowerCase()] !== "undefined" && CTE.translatedImageList[language.toLowerCase()].includes(imgSubfolder + newName)) {
				existsInList = true;
			}

			if (Utils.isNwjs()) { // Directly check if file exists
				var fs = require("fs");
				var path = require("path");
				var altImage = path.join(path.dirname(process.mainModule.filename), folder + "/" + newName + (!encrypted ? ".png" : ".rpgmvp"));
				if (fs.existsSync(altImage)) {
					if (!existsInList && CTE.MakeImageList) { // New file discovered; add it to list and re-save
						CTE.translatedImageList[language.toLowerCase()].push(imgSubfolder + newName);
						CTE.saveTranslatedImagesFile();
					}
					try {
						return CTE.ImageManager_loadBitmap.call(this, folder, newName, hue, smooth);
					}
					catch (e) {
					}
				}
			}
			else { // Refer to list of known translated images
				if (existsInList) {
					try {
						return CTE.ImageManager_loadBitmap.call(this, folder, newName, hue, smooth);
					}
					catch (e) {
					}
				}
			}
		}
	}

	return CTE.ImageManager_loadBitmap.apply(this, arguments);
};

// In playtest, checks image folders for language subfolders and generates TranslatedImage.txt.
CTE.generateTranslatedImageList = function() {
	if (!Utils.isNwjs()) return;

	var fs = require("fs");
	var path = require("path");
	var imgPath = path.join(path.dirname(process.mainModule.filename), "img/");

	var imageFolders = ["animations", "battlebacks1", "battlebacks2", "characters", "enemies", "faces", "parallaxes", "pictures", "sv_actors", "sv_enemies", "system", "tilesets", "titles1", "titles2"];
	for (var i = 0; i < imageFolders.length; i++) {
		for (var languageIndex = 0; languageIndex < CTE.LanguageCodes.length; languageIndex++) {
			var language = CTE.LanguageCodes[languageIndex].toLowerCase();
			var imageFolder = imageFolders[i];

			// Add any files in language subfolders.
			var subfolderPath = path.join(imgPath, imageFolders[i] + "/" + language);
			if (fs.existsSync(subfolderPath)) {
				var files = fs.readdirSync(subfolderPath);
				for (var fileIndex = 0; fileIndex < files.length; fileIndex++) {
					var file = files[fileIndex];
					if (file.toLowerCase().endsWith(".png") || file.toLowerCase().endsWith(".rpgmvp")) {
						if (file.toLowerCase().endsWith(".png")) file = file.slice(0, -4);
						else if (file.toLowerCase().endsWith(".rpgmvp")) file = file.slice(0, -7);
						CTE.translatedImageList[language] = CTE.translatedImageList[language] || [];
						CTE.translatedImageList[language].push(imageFolder + "/" + file);
					}
				}
			}
		}

		// Check all files in folder for ones ending in _XX (Base Language Code), and add equivalents that exist for each language.
		var folderPath = path.join(imgPath, imageFolders[i] + "/");
		if (fs.existsSync(folderPath)) {
			var files = fs.readdirSync(folderPath);
			for (var fileIndex = 0; fileIndex < files.length; fileIndex++) {
				var filename = files[fileIndex];
				if (filename.toLowerCase().endsWith(".png") || filename.toLowerCase().endsWith(".rpgmvp")) {
					if (filename.toLowerCase().endsWith(".png")) filename = filename.slice(0, -4);
					else if (filename.toLowerCase().endsWith(".rpgmvp")) filename = filename.slice(0, -7);
					if (filename.match(new RegExp("_" + CTE.BaseLanguage + "$"))) {
						for (var languageIndex = 0; languageIndex < CTE.LanguageCodes.length; languageIndex++) {
							var language = CTE.LanguageCodes[languageIndex];
							if (language === CTE.BaseLanguage) continue;

							var newName = filename.slice(0, -CTE.BaseLanguage.length) + language;
							if (fs.existsSync(folderPath + newName + ".png") || fs.existsSync(folderPath + newName + ".rpgmvp")) {
								language = language.toLowerCase();
								CTE.translatedImageList[language] = CTE.translatedImageList[language] || [];
								CTE.translatedImageList[language].push(imageFolder + "/" + newName);
							}
						}
					}
				}
			}
		}
	}

	CTE.saveTranslatedImagesFile();
};

// In playtest, save TranslatedImage.txt based on list data.
CTE.saveTranslatedImagesFile = function() {
	if (!Utils.isNwjs()) return;

	var output = "";
	var keys = Object.keys(CTE.translatedImageList);
	for (var i = 0; i < keys.length; i++) {
		output += "::" + keys[i].toUpperCase() + "\n";
		for (var k = 0; k < CTE.translatedImageList[keys[i]].length; k++) {
			output += CTE.translatedImageList[keys[i]][k] + "\n";
		}
		if (i < keys.length - 1) output += "\n";
	}

	if (output !== "") {
		try {
			var fs = require("fs");
			var path = require("path");
			var languagesPath = path.join(path.dirname(process.mainModule.filename), CTE_getLanguagesFolder());
			fs.writeFileSync(languagesPath + "TranslatedImages.txt", output.replace(/\n/g, "\r\n"));
		}
		catch (e) {
			CTE.warn("Failed to write TranslatedImages.txt. " + e.message);
		}
	}
};

// Outside of NW.js, load translated image list from TranslatedImages.txt.
CTE.loadTranslatedImagesFile = function() {
	var exists = CTE.load(CTE_getLanguagesFolder() + "TranslatedImages.txt");
	if (exists) {
		var language = "";
		var lines = CTE.fileText.replace(/\r/g, "").split("\n");

		for (var i = 0; i < lines.length; i++) {
			var line = lines[i].trim();
			if (line.startsWith("::")) {
				language = line.substring(2).toLowerCase();
			}
			else if (line !== "" && language !== "") {
				CTE.translatedImageList[language] = CTE.translatedImageList[language] || [];
				CTE.translatedImageList[language].push(line);
			}
		}
	}
};

/******************************
PAGE LANGUAGE CONDITIONS
******************************/

// Set up page language conditions whenever events refresh.
CTE.Game_Event_refresh = Game_Event.prototype.refresh;
Game_Event.prototype.refresh = function() {
	this.setPageLanguageConditions();
	CTE.Game_Event_refresh.apply(this, arguments);
};

// Define language conditions for each page based on whether there's a comment defining one at the top.
Game_Event.prototype.setPageLanguageConditions = function() {
	this.languageConditions = [];
	for (var i = 0; i < this.event().pages.length; i++) {
		var page = this.event().pages[i];

		var language = null;
		for (var commandIndex = 0; commandIndex < page.list.length; commandIndex++) {
			var command = page.list[commandIndex];
			if (!command) break;
			if (command.code === 108 || command.code === 408) {
				var comment = page.list[commandIndex].parameters[0];
				if (comment.toLowerCase().startsWith("<ctelanguage:")) {
					var startIndex = comment.toLowerCase().indexOf("<ctelanguage:") + 13;
					var endIndex = comment.indexOf(">", startIndex + 1);
					language = comment.substring(startIndex, endIndex).toLowerCase();
					break;
				}
			}
			else { // Stop as soon as a non-comment is reached
				break;
			}
		}

		if (language === null) { // No language
			this.languageConditions[i] = null;
		}
		else if (language.includes(",")) { // Multiple languages
			this.languageConditions[i] = language.split(",");
		}
		else { // Single language
			this.languageConditions[i] = [language];
		}
	}
};

// Alongside regular conditions, check language for pages that specify one, returning false if it doesn't match.
CTE.Game_Event_meetsConditions = Game_Event.prototype.meetsConditions;
Game_Event.prototype.meetsConditions = function(page) {
	var pageIndex = this.event().pages.indexOf(page);
	var condition = this.languageConditions[pageIndex];
	if (condition !== null && !condition.includes(ConfigManager.getLanguage().toLowerCase())) {
		return false;
	}
	return CTE.Game_Event_meetsConditions.apply(this, arguments);
};

/******************************
CONFIG LANGUAGE SETTING
******************************/

ConfigManager._language = CTE.DefaultLanguage !== "" ? CTE.DefaultLanguage : CTE.BaseLanguage;

Object.defineProperty(ConfigManager, "language", {
	get: function() {
		return this._language;
	},
	set: function(value) {
		this._language = value;
		Scene_Boot.prototype.updateDocumentTitle.call();

		// Update anything custom.
		CTE.languageRefreshHandler();
	},
	configurable: true
});

ConfigManager.getLanguage = function() {
	return this._language;
};

ConfigManager.getLanguageIndex = function() {
	return CTE.LanguageCodes.indexOf(this.getLanguage());
};

ConfigManager.isLanguage = function(language) {
	return this.getLanguage().toLowerCase() === language.toLowerCase();
};

ConfigManager.isBaseLanguage = function() {
	return this.isLanguage(CTE.BaseLanguage);
};

CTE.ConfigManager_makeData = ConfigManager.makeData;
ConfigManager.makeData = function() {
	const config = CTE.ConfigManager_makeData.apply(this, arguments);
	config.language = this.language;
	return config;
};

CTE.ConfigManager_applyData = ConfigManager.applyData;
ConfigManager.applyData = function(config) {
	CTE.ConfigManager_applyData.apply(this, arguments);
	this.language = this.readLanguage(config, "language");
};

ConfigManager.readLanguage = function(config, name) {
	if (CTE.LockLanguage) return CTE.DefaultLanguage;
	if (config[name]) {
		return config[name];
	} else {
		const userLanguageCode = CTE.languageCodeFromExternalAPI();
		return (userLanguageCode !== "" ? userLanguageCode : CTE.DefaultLanguage !== "" ? CTE.DefaultLanguage : CTE.BaseLanguage);
	}
};

/******************************
ADD LANGUAGE OPTION
******************************/

if (!CTE.LockLanguage) {
	CTE.Window_Options_addGeneralOptions = Window_Options.prototype.addGeneralOptions;
	Window_Options.prototype.addGeneralOptions = function() {
		CTE.Window_Options_addGeneralOptions.apply(this, arguments);
		this.addCommand(CTE.getString("LanguageOptionName", CTE.BaseLanguageOptionName), "language");
	};

	CTE.Window_Options_statusText = Window_Options.prototype.statusText;
	Window_Options.prototype.statusText = function(index) {
		const symbol = this.commandSymbol(index);
		if (symbol === "language") {
			const value = this.getConfigValue(symbol);
			if (value.toLowerCase() === CTE.BaseLanguage.toLowerCase()) {
				return CTE.BaseLanguageDisplayName;
			} else if (CTE.dataExistsForLanguage(value, "extra", "LanguageDisplayName")) {
				const displayName = CTE.getDataForLanguage(value, "extra", "LanguageDisplayName");
				return displayName !== "" ? displayName : value; // Default to language code if left blank
			} else {
				return value;
			}
		}
		return CTE.Window_Options_statusText.apply(this, arguments);
	};

	Window_Options.prototype.changeLanguage = function(symbol, value, change) {
		let index = CTE.LanguageCodes.indexOf(value);
		index = (index + CTE.LanguageCodes.length + change) % CTE.LanguageCodes.length;
		value = CTE.LanguageCodes[index];
		this.changeValue(symbol, value);
		this.refresh();
	};

	CTE.createCheckLanguageChange = function(delegate, change) {
		return function() {
			const index = this.index();
			const symbol = this.commandSymbol(index);
			if (symbol === "language") {
				const value = this.getConfigValue(symbol);
				return this.changeLanguage(symbol, value, change);
			}
			return delegate.apply(this, arguments);
		}
	};

	CTE.Window_Options_processOk = Window_Options.prototype.processOk;
	Window_Options.prototype.processOk = CTE.createCheckLanguageChange(CTE.Window_Options_processOk, 1);

	CTE.Window_Options_cursorRight = Window_Options.prototype.cursorRight;
	Window_Options.prototype.cursorRight = CTE.createCheckLanguageChange(CTE.Window_Options_cursorRight, 1);

	CTE.Window_Options_cursorLeft = Window_Options.prototype.cursorLeft;
	Window_Options.prototype.cursorLeft = CTE.createCheckLanguageChange(CTE.Window_Options_cursorLeft, -1);
}

/******************************
USE DEFAULT FONT FOR LANGUAGE
******************************/

// Preload defined default font for language as soon as language data has been loaded.
CTE.preloadDefaultFont = function(language = null) {
	if (language === null) language = ConfigManager.getLanguage();

	CTE.loadFont(language, "DefaultFont");
	CTE.loadFont(language, "NumberFont");
};

CTE.loadFont = function(language, font) {
	if (CTE.dataExistsForLanguage(language, "extra", font)) {
		const myFont = CTE.getDataForLanguage(language, "extra", font);
		if (myFont !== "" && !CTE.preloadedFonts.contains(myFont)) {
			FontManager.load(myFont, myFont);
			CTE.preloadedFonts.push(myFont);
		}
	}
}

// Use specified default font for language.
CTE.Game_System_mainFontFace = Game_System.prototype.mainFontFace;
Game_System.prototype.mainFontFace = function() {
	const myFont = CTE.getString("DefaultFont", "");
	return myFont !== "" ? myFont : CTE.Game_System_mainFontFace.apply(this, arguments);
};

// Use specified default font for language.
CTE.Game_System_numberFontFace = Game_System.prototype.numberFontFace;
Game_System.prototype.numberFontFace = function() {
	const myFont = CTE.getString("NumberFont", "");
	return myFont !== "" ? myFont : CTE.Game_System_numberFontFace.apply(this, arguments);
};

// Use specified default font size offset for language.
CTE.Game_System_mainFontSize = Game_System.prototype.mainFontSize;
Game_System.prototype.mainFontSize = function() {
	const myFontSizeOffset = CTE.getNumber("DefaultFontSizeOffset", 0);
	return CTE.Game_System_mainFontSize.apply(this, arguments) + myFontSizeOffset;
};

/******************************
MESSAGE SPEED
******************************/

// Initialize speed modifier variables when a message starts.
CTE.Window_Message_startMessage = Window_Message.prototype.startMessage;
Window_Message.prototype.startMessage = function() {
	CTE.Window_Message_startMessage.apply(this, arguments);
	this.speedBoostTimer = 0;
	this.extraCharCount = 0;
};

// Modify number of characters typed every X frames based on message speed setting for language.
CTE.Window_Message_updateMessage = Window_Message.prototype.updateMessage;
Window_Message.prototype.updateMessage = function() {
	if (Object.keys(PluginManager.parameters("MessageSpeedCustomize")).length > 0) { // Incorporate speed handling from MessageSpeedCustomize if using it
		var speed = this.getMessageSpeed();
		if (this._textState && !this._lineShowFast) {
			if (speed <= 0 || this._showAll) {
				this._showFast = true;
			} else if (!this.isEndOfText(this._textState)) {
				this._waitCount = speed - 1;
			}
		}
	}

	var speedSetting = CTE.getString("MessageSpeed", "1");
	if (speedSetting !== "" && speedSetting !== "1") {
		speedSetting = speedSetting.replace(/x/gi, "").replace(/0\./g, ".");
	}

	var modifier = 0
	var modifyPeriod = 0;
	switch (speedSetting) {
		case ".125": modifier = -7; modifyPeriod = 1; break; // 1/8
		case ".25": modifier = -3; modifyPeriod = 1; break; // 1/4
		case ".5": modifier = -1; modifyPeriod = 1; break; // 1/2
		case ".66": modifier = -1; modifyPeriod = 2; break; // 2/3 (repeating)
		case ".75": modifier = -1; modifyPeriod = 3; break; // 3/4
		case ".8": modifier = -1; modifyPeriod = 4; break; // 4/5
		case ".9": modifier = -1; modifyPeriod = 9; break; // 10/11 (approximate)
		case "1.1": modifier = 1; modifyPeriod = 10; break; // 11/10
		case "1.125": modifier = 1; modifyPeriod = 8; break; // 9/8
		case "1.16": modifier = 1; modifyPeriod = 6; break; // 7/6 (repeating)
		case "1.2": modifier = 1; modifyPeriod = 5; break; // 6/5
		case "1.25": modifier = 1; modifyPeriod = 4; break; // 5/4
		case "1.33": modifier = 1; modifyPeriod = 3; break; // 4/3 (repeating)
		case "1.5": modifier = 1; modifyPeriod = 2; break; // 3/2
		case "1.66": modifier = 1; modifyPeriod = 2; break; // 5/3 (repeating)
		case "2": modifier = 1; modifyPeriod = 1; break; // 2/1
	}

	// At 1x message speed, just use original function.
	if (modifier == 0) return CTE.Window_Message_updateMessage.apply(this, arguments);

	if (this._textState) {
		while (!this.isEndOfText(this._textState)) {
			if (this.needsNewPage(this._textState)) {
				this.newPage(this._textState);
			}

			if (this.extraCharCount < 0) { // Skip this frame if working through a negative modifier
				this.extraCharCount++;
				break;
			}

			this.updateShowFast();
			this.processCharacter(this._textState);

			if (this.extraCharCount <= 0) { // Advance loop timer if not working through extra characters
				this.speedBoostTimer++;
			}
			else { // Otherwise, work through extra characters
				this.extraCharCount--;
			}

			if (this.speedBoostTimer >= modifyPeriod) { // When loop timer reaches end of period, set the character modifier
				this.speedBoostTimer = 0;
				this.extraCharCount = modifier;
			}

			if (this.pause || this._waitCount > 0) {
				if (this.extraCharCount > 0) this.extraCharCount = 0; // If paused by a control code, give up on typing extra characters
				break;
			}
			if (!this._showFast && !this._lineShowFast && this.extraCharCount <= 0) { // Barring instant-type mode, stop processing characters if dealing with a negative modifier
				break;
			}
		}
		if (this.isEndOfText(this._textState)) {
			this.onEndOfText();
		}
		return true;
	} else {
		return false;
	}
};

/******************************
TEXT/CSV EXPORTING
******************************/

// Begin processing base data for export.
CTE.beginExportProcess = function() {
	if (CTE.exportingBase) {
		CTE.warn("Exporting text data for base language (" + (CTE.BaseExportFilename || CTE.BaseLanguage) + ")...");
	}
	if (CTE.exportingIdentical) {
		CTE.warn("Checking for identical text (logged to IdenticalWarnings.txt)...");
	}
	if (CTE.exportingReformat) {
		CTE.warn("Exporting reformatted text to " + (CTE.languageForReformat) + " (Reformatted)...");
	}

	CTE.exportedContent = { text: "", table: "" };
	CTE.exportedIdentical = "";
	CTE.variableStringAssignments = {};
	CTE.speakerNameList = [];

	if (CTE.exportingIdentical || CTE.exportingReformat) {
		CTE.orderedLanguages = [...CTE.LanguageCodes];
		if (CTE.orderedLanguages.length > 0 && CTE.orderedLanguages[0].toLowerCase() !== CTE.BaseLanguage.toLowerCase()) {
			var baseIndex = CTE.orderedLanguages.indexOf(CTE.BaseLanguage);
			if (baseIndex != -1) {
				CTE.orderedLanguages.splice(baseIndex, 1);
				CTE.orderedLanguages.unshift(CTE.BaseLanguage);
			}
		}
	}

	CTE.exportingPosition = {};

	// Start by exporting all maps.
	CTE.exportingMapNum = 1;
	CTE.loadNextMap();
};

// Check for database loaded status to perform launch actions once ready. Otherwise, call CTE.onDataLoad to process certain data files.
CTE.DataManager_onLoad = DataManager.onLoad;
DataManager.onLoad = function(object) {
	if (object === $CTETemp) { // After temp-loading a map, export text from it, then move to next map (or databases, )
		CTE.exportMapText();
		CTE.exportingMapNum++;
		if (CTE.exportingMapNum < $dataMapInfos.length) {
			CTE.loadNextMap();
		}
		else {
			CTE.exportPlugins();
			CTE.exportDatabases();
		}
	}
	else {
		// Create translated image list, either by checking files directly in NW.js or by loading from list.
		if (CTE.launchImageListPending) {
			CTE.launchImageListPending = false;
			if (Utils.isNwjs() && CTE.isPlaytest() && CTE.MakeImageList) { // Generate list
				CTE.generateTranslatedImageList();
			}
			else if (!Utils.isNwjs()) { // Load list from file
				CTE.loadTranslatedImagesFile();
			}
		}

		// Once enough of database is loaded to do so, load language scripts.
		if (CTE.launchLoadPending && $dataMapInfos && $dataCommonEvents && $dataTroops) {
			CTE.launchLoadPending = false;
			CTE.loadLanguageScripts();
		}

		// If database is done loading and launch export has not been performed, start the process.
		if (CTE.launchExportPending && DataManager.isDatabaseLoaded()) {
			CTE.launchExportPending = false;
			if (Utils.isNwjs() && CTE.isPlaytest()) {
				if (CTE.ExportBaseOnLaunch || CTE.ExportTableOnLaunch || CTE.ExportIdenticalOnLaunch) {
					CTE.exportingBase = CTE.ExportBaseOnLaunch;
					CTE.exportingIdentical = CTE.ExportIdenticalOnLaunch;
					CTE.exportingReformat = false;
					CTE.beginExportProcess();
				}
			}
		}

		// Specially process certain files for in-game purposes.
		CTE.onDataLoad(object);

		// Go to original function.
		CTE.DataManager_onLoad.apply(this, arguments);
	}
};

// Load map file (if it exists) and wait for it to load. Move on to the next if it doesn't exist.
CTE.loadNextMap = function() {
	if (!Utils.isNwjs()) return;

	const filename = "Map%1.json".format(CTE.exportingMapNum.padZero(3));
	if ($dataMapInfos[CTE.exportingMapNum]) {
		$CTETemp = {};
		DataManager.loadDataFile("$CTETemp", filename);
	}
	else {
		CTE.exportingMapNum++;
		if (CTE.exportingMapNum < $dataMapInfos.length) {
			CTE.loadNextMap();
		} else {
			CTE.exportPlugins();
			CTE.exportDatabases();
		}
	}
};

// Export map text from $CTETemp.
CTE.exportMapText = function() {
	const exportEntries = CTE.createTranslationData($CTETemp, ["displayName", "list"], "", { "list": CTE.exportCommandList });
	CTE.exportedContent.text = JSON.stringify(exportEntries, null, 4);

	if (!CTE.AllInOneFile && !CTE.AllInOneMaps) {
		CTE.writeBaseFile("Map%1".format(CTE.exportingMapNum.padZero(3)));
	}
}

CTE.writeCommandContent = function(command, content, commandCounter, commandName, str, key = null) {
	const commandContent = { text: "" };
	commandCounter[commandName] = (commandCounter[commandName] || 0) + 1;

	CTE.writeFormat(commandContent, "{0}", str, function(language) {
		const commandText = key ? CTE.getDataForLanguage(language, key, str) : CTE.getCommandTextForCSV(language, commandName, commandCounter);
		if (commandText !== null && !commandText.includes("@if") && str.split("\n").length !== commandText.split("\n").length) {
			CTE.warn(commandName + " in " + language + " do not match base in " + CTE.exportingPosition, 2);
		}
		return commandText;
	}, `${CTE.exportingPosition.file}-${CTE.exportingPosition.event}-${CTE.exportingPosition.page}-${commandCounter[commandName]}`, commandName);

	const customPrefix = CTE.commandExportHandlerPrefix(command);
	if (customPrefix !== "") commandContent.text = customPrefix + commandContent.text
	const customPostfix = CTE.commandExportHandlerPostfix(command);
	if (customPostfix !== "") commandContent.text += customPostfix;

	content[`${commandName.toLowerCase()}-${commandCounter[commandName]}`] = commandContent.text;
};

CTE.processMultiLineCommand = function(command, content, commandCounter, list, commandNum, code, commandName) {
	const lines = [];
	while (list[++commandNum]?.code === code) {
		lines.push(list[commandNum].parameters[0]);
	}

	commandCounter.multiLine = lines.length;
	CTE.writeCommandContent(command, content, commandCounter, commandName, lines.join("\n"));
};

CTE.handleTextCommand = function(command, content, commandCounter, list, commandNum) {
	const speakerName = command.parameters[4] || "";
	if (speakerName !== "" && !CTE.speakerNameList.contains(speakerName)) {
		CTE.speakerNameList.push(speakerName);
	}
	CTE.processMultiLineCommand(command, content, commandCounter, list, commandNum, 401, "Message");
};

CTE.handleScrollingTextCommand = function(command, content, commandCounter, list, commandNum) {
	CTE.processMultiLineCommand(command, content, commandCounter, list, commandNum, 405, "ScrollingMessage");
};

CTE.handleChoiceCommand = function(command, content, commandCounter) {
	const str = command.parameters[0].join("\n");
	CTE.writeCommandContent(command, content, commandCounter, "Choice", str);
};

CTE.handleVariableScript = function(command) {
	if (command.parameters[3] !== 4) return false; // Not a script
	const script = command.parameters[4].trim();
	try {
		if ((script.startsWith("\"") && script.endsWith("\"")) || (script.startsWith("'")) && script.endsWith("'") || (script.startsWith("`")) && script.endsWith("`")) {
			const value = eval(script);
			if (typeof value === "string") {
				for (let i = command.parameters[0]; i <= command.parameters[1]; i++) {
					CTE.variableStringAssignments[i] = CTE.variableStringAssignments[i] || {};
					CTE.variableStringAssignments[i][value] = value;
					j++;
				}
			}
		}
	} catch (e) { }
};

CTE.handleChangeTextCommand = function(commandName) {
	return (command, content, commandCounter) => {
		const originalString = command.parameters[1].replace(/\n/g, "\\n");
		return CTE.writeCommandContent(command, content, commandCounter, commandName, originalString, commandName);
	};
};

CTE.exportCommandListHandlers = {
	101: CTE.handleTextCommand,
	102: CTE.handleChoiceCommand,
	105: CTE.handleScrollingTextCommand,
	122: CTE.handleVariableScript,
	320: CTE.handleChangeTextCommand("namechange"),
	324: CTE.handleChangeTextCommand("nicknamechange"),
	325: CTE.handleChangeTextCommand("profilechange")
};

// Export text for a list of commands in a page, common event, or troop page.
CTE.exportCommandList = function(content, list) {
	const commandCounter = {};
	for (let commandNum = 0; commandNum < list.length; commandNum++) {
		const command = list[commandNum];
		const commandHandler = CTE.exportCommandListHandlers[command.code];
		if (commandHandler) {
			commandHandler(command, content, commandCounter, list, commandNum);
			if (commandCounter.multiLine) {
				commandNum += commandCounter.multiLine;
				commandCounter.multiLine = 0;
			}
		}
	}
};

CTE.exportDatabase = function(name, database) {
	CTE.exportingPosition["file"] = name;
	const data = database.data || window["$data" + name]
	const exportEntries = CTE.createTranslationData(data, database.rules || [""], (database?.prefix + "-") || "", database.handlers || {});

	if (!CTE.AllInOneFile && CTE["AllInOne" + name] === false) {
		const capitalizedPrefix = database.prefix.charAt(0).toUpperCase() + database.prefix.slice(1);
		const fileName = capitalizedPrefix + "%1";
		for (let i = 1; i <= data.length; i++) {
			const content = exportEntries[database.prefix + "-" + i] || {};
			const id = content.id || i;
			CTE.exportedContent.text = JSON.stringify(content, null, 4);
			CTE.writeBaseFile(fileName.format(id.padZero(4)));
			CTE.exportedContent.text = "";
		}
	} else {
		CTE.exportedContent.text = JSON.stringify(exportEntries, null, 4);
	}
}

CTE.createDatabaseMap = function() {
	CTE.databaseMap = {
		"Commons": { data: () => $dataCommonEvents, rules: ["list"], prefix: "common", handlers: { "list": CTE.exportCommandList } },
		"Troops": { rules: ["list"], prefix: "troop", handlers: { "list": CTE.exportCommandList } },
		"Actors": { rules: ["name", "nickname", "profile"], prefix: "actor" },
		"Skills": { rules: ["name", "description", "message1", "message2", "note"], prefix: "skill", updateDatabaseProperties: true },
		"Items": { rules: ["name", "description", "note"], prefix: "item", updateDatabaseProperties: true },
		"Weapons": { rules: ["name", "description"], prefix: "weapon", updateDatabaseProperties: true },
		"Armors": { rules: ["name", "description"], prefix: "armor", updateDatabaseProperties: true },
		"States": { rules: ["name", "message1", "message2", "message3", "message4"], prefix: "state", updateDatabaseProperties: true },
		"Classes": { rules: ["name"], prefix: "class", updateDatabaseProperties: true },
		"Enemies": { rules: ["name"], prefix: "enemy" },
		"System": { rules: ["gameTitle", "currencyUnit", "skillTypes", "equipTypes", "terms"] },
		"Variables": { data: () => CTE.variableStringAssignments },
		"Speakers": { data: () => CTE.speakerNameList, handlers: { "__array": (exportEntries, entry) => exportEntries[entry] = entry } },
		"Extra": { data: () => CTE.extraSectionBase }
	};

	// Add in databases from plugins
	CTE.PluginDatabases.forEach(db => CTE.databaseMap[db] = { rules: CTE.PluginParamsToExport });
}

CTE.getDatabases = function() {
	if (!CTE.databaseMap) {
		CTE.createDatabaseMap();
	}

	return CTE.databaseMap;
}

// After exporting maps, export the rest from databases.
CTE.exportDatabases = function() {
	const allDatabases = CTE.getDatabases();

	if (!CTE.AllInOneFile && CTE.AllInOneMaps) { // Export map text to combined map file is enabled
		CTE.writeBaseFile("AllMaps");
	}

	for (const [name, database] of Object.entries(allDatabases)) {
		if (typeof database.data == "function") database.data = database.data();
		CTE.exportDatabase(name, database);
		if (CTE["AllInOne" + name] === undefined) {
			if (!CTE.AllInOneFile && !CTE.AllInOneDatabase) {
				CTE.writeBaseFile(name);
			}
		} else {
			if (!CTE.AllInOneFile && CTE["AllInOne" + name]) {
				CTE.writeBaseFile("All" + name);
			}
		}
	}

	if (!CTE.AllInOneFile && CTE.AllInOneDatabase) {
		CTE.writeBaseFile("AllDatabases");
	}

	CTE.exportFinish();
};

// Generic functionality to extract translation data from a dataset
// data: Data from which translation data should be extracted
// rules: Only data properties that have a suffix in rules will be extracted
// prefix: Optional top-level prefix that will be added to data keys
// customDataHandlers: Data extraction will either proceed generically, or use a custom handler for data if defined for a property
CTE.createTranslationData = function(data, rules, prefix = "", customDataHandlers = {}, parent = "") {
	const exportEntries = {};

	if (Array.isArray(data)) {
		for (let i = 0; i < data.length; i++) {
			if (customDataHandlers["__array"]) {
				// The special key __array indicates we want to use a custom handler for arrays
				customDataHandlers["__array"](exportEntries, data[i], i);
				continue;
			}

			if (data[i] == null) continue;
			if (typeof data[i] == "string") {
				if (data[i].length == 0) continue;
				// For arrays of strings, only assign data if the parent node fulfills a rule
				if (parent != "" && !rules.some(rule => parent.endsWith(rule))) continue;
				exportEntries[prefix + i] = data[i];
			} else {
				const translationData = CTE.createTranslationData(data[i], rules, "", customDataHandlers, parent);
				// Check if the data is empty
				// Possible if there were no sub objects that fit the rules
				if (Object.keys(translationData).length == 0) continue;
				exportEntries[prefix + i] = translationData;
			}
		}
	} else if (data != null && typeof data == "object") {
		// Sort keys according to rule priority 
		// This way keys such as "name" can be put at the start of the json file
		const keys = Object.keys(data)
			.sort((key1, key2) =>
				(rules.includes(key1) ? rules.indexOf(key1) : Infinity)
				- (rules.includes(key2) ? rules.indexOf(key2) : Infinity));

		for (const key of keys) {
			const entry = data[key];
			// Either proceed generically or call a custom handler if one is defined
			if (customDataHandlers[key]) {
				customDataHandlers[key](exportEntries, entry, key);
				continue;
			}

			const fulfillsRule = rules.some(rule => key.endsWith(rule));
			if (typeof entry == "string") {
				if (entry === null || entry.length === 0) continue;
				// If the key doesn't fulfll a rule, skip
				if (!fulfillsRule) continue;
				// For notetags, we also check if there is a notetag that should be exported
				if (key == "note" && !CTE.NotetagsToExport.some(tag => data.meta[tag])) continue;
				exportEntries[key] = entry;
			} else if (entry != null && typeof entry == "object") {
				// If a parent fulfills a rule, then all children fulfill it
				const subRules = fulfillsRule ? [""] : rules;
				const translationData = CTE.createTranslationData(entry, subRules, "", customDataHandlers, key);
				if (Object.keys(translationData).length == 0) continue;
				exportEntries[key] = translationData;
			}
		}
	}

	return exportEntries;
}

// Generic export for translation data
CTE.exportTranslationData = function(data, rules, prefix, customDataHandlers = {}) {
	const exportEntries = CTE.createTranslationData(data, rules, prefix, customDataHandlers);
	CTE.exportedContent.text = JSON.stringify(exportEntries, null, 4);
}

// Export text from $plugins.
CTE.exportPlugins = function() {
	const exportEntries = {
		"plugins": CTE.createTranslationData($plugins, [], "", {
			"__array": (exportEntries, entry) => {
				const pluginParams = {};
				TLB.SKABase.parseParameters(entry.parameters, pluginParams);
				const translationData = CTE.createTranslationData(pluginParams, CTE.PluginParamsToExport);
				if (Object.keys(translationData).length == 0) return;
				exportEntries[entry.name] = translationData;
			}
		})
	};

	CTE.exportedContent.text = JSON.stringify(exportEntries, null, 4);
	if (!CTE.AllInOneFile) {
		CTE.writeBaseFile("AllPlugins");
	}
}

// Add the Extra section at the end, for defining arbitrary data values.
CTE.extraSectionBase = {
	LanguageDisplayName: CTE.BaseLanguageDisplayName,
	LanguageOptionName: CTE.BaseLanguageOptionName,
	DefaultFont: "GameFont",
	DefaultFontSizeOffset: "0",
	MessageSpeed: CTE.BaseMessageSpeed
}

// Write to text/table following a format, using baseValue for base language and languageFunction(language) for others. Returns boolean for whether it wrote something.
CTE.writeFormat = function(content, format, baseValue, languageFunction, locationID = "", fieldName = "", missingWarning = true) {
	if (CTE.exportingBase) {
		content.text += format.replace(/\{0\}/g, baseValue).replace(/\{1\}/g, "");
	}

	if (CTE.exportingReformat) {
		var language = CTE.languageForReformat;
		var myValue = languageFunction(language);
		var ignoreMe = false;
		if (myValue === null) { // Language function didn't find anything
			myValue = "<<<NOT FOUND>>>";
			if (CTE.shouldIgnore(language, locationID, fieldName)) {
				return false;
			}
			else if (missingWarning) {
				CTE.warn(language + " text not found for " + CTE.exportingPosition, 2);
			}
		}
		var comment = "";
		if (CTE.shouldIgnore(language, locationID, fieldName, true)) comment = " %ignorethis"; // Specifically ignored
		if (baseValue === "<<<NOT FOUND IN BASE>>>") { // Excess text
			comment = " " + baseValue;
			format = format.replace(/#([^#]+)#/, function(match, commandName) { return "_" + commandName + "_"; });
		}
		content.text += format.replace(/\{0\}/g, myValue).replace(/\{1\}/g, comment);
	}

	return true;
};

// Write exported text to file.
CTE.exportFinish = function() {
	if (!Utils.isNwjs()) return;

	const fs = require("fs");
	const path = require("path");
	const languagesPath = path.join(path.dirname(process.mainModule.filename), CTE_getLanguagesFolder());

	try {
		if (!fs.existsSync(languagesPath)) {
			fs.mkdirSync(languagesPath);
		}

		if (CTE.AllInOneFile && (CTE.exportingBase || CTE.exportingReformat)) {
			CTE.writeBaseFile();
		}

		if (CTE.exportingBase) {
			CTE.warn("Finished exporting text data for base language (" + (CTE.BaseExportFilename || CTE.BaseLanguage) + ").");
		}

		if (CTE.exportingReformat) {
			CTE.warn("Finished exporting reformatted text to " + CTE.languageForReformat + " (Reformatted).");
		}

		if (CTE.exportingIdentical) {
			try {
				if (CTE.exportedIdentical !== "") {
					CTE.warn("Wrote log of identical text (IdenticalWarnings.txt).");
				} else {
					CTE.warn("No identical text found between languages.");
				}
				const filename = "IdenticalWarnings.txt";
				if (CTE.exportedIdentical !== "" || fs.existsSync(languagesPath + filename)) { // Overwrite file with blank one if a log already exists
					fs.writeFileSync(languagesPath + filename, CTE.exportedIdentical.replace(/\n/g, "\r\n"));
				}
			}
			catch (e) {
				CTE.warn("Failed to write " + filename + ". " + e.message);
			}
		}
	}
	catch (e) {
		CTE.warn("Failed to export. " + e.message);
	}

	CTE.exportingBase = false;
	CTE.exportingIdentical = false;
	CTE.exportingReformat = false;
};

// Write a file for the base script, either the entire thing all in one or an individual part in the subfolder for that language.
CTE.writeBaseFile = function(type = "") {
	if (!Utils.isNwjs()) {
		CTE.exportedContent.text = "";
		return;
	}

	if (!CTE.exportingBase && !CTE.exportingReformat) return;

	const fs = require("fs");
	const path = require("path");
	const languagesPath = path.join(path.dirname(process.mainModule.filename), CTE_getLanguagesFolder());
	const language = CTE.exportingBase ? (CTE.BaseExportFilename || CTE.BaseLanguage) : `${CTE.languageForReformat} (Reformatted)`;

	let [folder, filename] = type.includes("/") ? type.split("/") : ["", type];
	filename = (filename || language) + ".json";

	const targetFolder = path.join(languagesPath, language, folder);

	try {
		fs.mkdirSync(targetFolder, { recursive: true });
		fs.writeFileSync(path.join(targetFolder, filename), CTE.exportedContent.text.replace(/\n/g, "\r\n"));
	} catch (e) {
		CTE.warn(`Failed to write ${filename}. ${e.message}`);
	}

	CTE.exportedContent.text = "";
};

/******************************
MESSAGE DEBUGGER
******************************/

CTE.TextDebug = CTE.TextDebug || {};
CTE.TextDebug.currentFile = "";
CTE.TextDebug.currentEvent = -1;
CTE.TextDebug.lastFileIndex = 0;
CTE.TextDebug.lastIndex = new Map();
CTE.TextDebug.searchString = "";

// Return identifier for storing lastIndex values.
CTE.TextDebug.getFileIdentifier = function() {
	return this.currentEvent == -1 ? this.currentFile : this.currentFile + this.currentEvent
};

// Message debugger container scene.
function Scene_TextDebugger() {
	this.initialize.apply(this, arguments);
};

Scene_TextDebugger.prototype = Object.create(Scene_MenuBase.prototype);
Scene_TextDebugger.prototype.constructor = Scene_MenuBase;

Scene_TextDebugger.prototype.initialize = function() {
	Scene_MenuBase.prototype.initialize.call(this);
	this._choiceWindow = new TextDebugger_Window_ChoiceList();
};

Scene_TextDebugger.prototype.create = function() {
	Scene_MenuBase.prototype.create.call(this);
	this.removeChild(this._choiceWindow);
	this.addChild(this._choiceWindow);
	this._backgroundSprite.visible = false;
};

// Message debugger menu window.
function TextDebugger_Window_ChoiceList() {
	this.initialize.apply(this, arguments);
};

TextDebugger_Window_ChoiceList.prototype = Object.create(Window_Selectable.prototype);
TextDebugger_Window_ChoiceList.prototype.constructor = TextDebugger_Window_ChoiceList;

// Initialization.
TextDebugger_Window_ChoiceList.prototype.initialize = function() {
	this._list = [];
	Window_Selectable.prototype.initialize.call(this, new Rectangle(0, 0, Graphics.width, Graphics.height));
	this.openness = 255;
	this.active = true;
	this.updatePlacement();
	this.makeCommandList();
	this.select(CTE.TextDebug.currentFile === "" ? CTE.TextDebug.lastFileIndex : CTE.TextDebug.lastIndex.get(CTE.TextDebug.getFileIdentifier()));
	this.refresh();
};

// Set up window position and size.
TextDebugger_Window_ChoiceList.prototype.updatePlacement = function() {
	this.x = 0;
	this.y = 0;
	this.width = Graphics.width;
	this.height = Graphics.height;
};

// Generate choice list for either all files, or messages within a file.
TextDebugger_Window_ChoiceList.prototype.makeCommandList = function() {
	this._list = [];
	if (CTE.TextDebug.currentFile === "") { // File list
		this.makeFileList();
	}
	else if (CTE.TextDebug.currentFile == "commonevents") { // Common event message list
		this.makeCommonMessageList(CTE.TextDebug.currentEvent);
	}
	else if (CTE.TextDebug.currentFile == "troops") { // Troop message list
		this.makeTroopMessageList(CTE.TextDebug.currentEvent);
	}
	else { // Map message list
		this.makeMapMessageList(CTE.TextDebug.currentFile);
	}
};

// Make file list out of data for current language.
TextDebugger_Window_ChoiceList.prototype.makeFileList = function() {
	var language = ConfigManager.getLanguage();
	if (typeof CTE.data[language] === "undefined") {
		if (ConfigManager.isBaseLanguage()) {
			Window_Command.prototype.addCommand.call(this, "No language data for Base Language.", "");
		}
		else {
			Window_Command.prototype.addCommand.call(this, "Language data not found for " + language + ".", "");
		}
		return;
	}

	for (const fileKey of Object.keys(CTE.data[language])) {
		if (fileKey.startsWith("map")) {
			const mapData = CTE.data[language][fileKey];
			const hasMessages = !CTE.traverseLeaves(mapData, null, null, (_, key, __) => !key.contains("message-"));
			if (hasMessages) {
				const mapID = fileKey.replace("map", "");
				const mapName = $dataMapInfos[mapID].name || "";
				const previewText = "Map " + mapID + (mapName !== "" ? ": " + mapName : "");
				Window_Command.prototype.addCommand.call(this, previewText, fileKey);
			}
		}
		else if (fileKey.startsWith("commonevents")) {
			var commonKeys = Object.keys(CTE.data[language][fileKey]);
			var commonList = [];

			for (var k = 0; k < commonKeys.length; k++) {
				var dashIndex = commonKeys[k].indexOf("-");
				var commonID = commonKeys[k].substring(0, dashIndex);
				if (dashIndex != -1 && commonKeys[k].contains("-message") && !commonList.includes(commonID)) {
					commonList.push(commonID);
				}
			}

			for (var k = 0; k < commonList.length; k++) {
				var commonID = commonList[k];
				var myData = $dataCommonEvents;
				var indexInData = commonID;
				var commonName = myData[indexInData].name || "";
				var previewText = "Common Event " + commonID + (commonName !== "" ? ": " + commonName : "");
				Window_Command.prototype.addCommand.call(this, previewText, "common" + commonID);
			}
		}
		else if (fileKey.startsWith("troops")) {
			var troopKeys = Object.keys(CTE.data[language][fileKey]);
			var troopList = [];

			for (var k = 0; k < troopKeys.length; k++) {
				var dashIndex = troopKeys[k].indexOf("-");
				var troopID = troopKeys[k].substring(0, dashIndex);
				if (dashIndex != -1 && troopKeys[k].contains("-message") && !troopList.includes(troopID)) {
					troopList.push(troopID);
				}
			}

			for (var k = 0; k < troopList.length; k++) {
				var troopID = commonList[k];
				var troopName = $dataTroops[troopID].name || "";
				var previewText = "Troop " + troopID + (troopName !== "" ? ": " + troopName : "");
				Window_Command.prototype.addCommand.call(this, previewText, "troop" + troopID);
			}
		}
	}
};

// Make list of messages in map, with unselectable headers for event and page.
TextDebugger_Window_ChoiceList.prototype.makeMapMessageList = function(file) {
	const language = ConfigManager.getLanguage();
	CTE.loadLanguage(language);

	const events = CTE.getDataForLanguage(language, file, "events");
	let lastHeader = "";
	for (const event in events) {
		const eventID = parseInt(event);
		const pages = events[event]["pages"];
		for (const page in pages) {
			const pageID = parseInt(page);
			const list = pages[page];
			for (const dataKey in list) {
				const header = "\\c[1]***** Event " + eventID + ", Page " + pageID + " *****";
				if (header !== lastHeader) {
					Window_Command.prototype.addCommand.call(this, header, "");
					lastHeader = header;
				}

				if (!dataKey.contains("message-")) continue;
				const message = list[dataKey];
				const displayText = this.formatDisplayMessage(message);
				const messageData = {};
				messageData.eventID = eventID;
				messageData.pageID = pageID;
				messageData.messageID = dataKey;
				Window_Command.prototype.addCommand.call(this, displayText, messageData);
			}
		}
	}
};

// Make list of messages in common event.
TextDebugger_Window_ChoiceList.prototype.makeCommonMessageList = function(event) {
	CTE.loadLanguage(ConfigManager.getLanguage());

	var dataKey = event + "-0-message";
	var messages = CTE.getData("commonevents", dataKey);
	if (messages === null) return;

	for (var i = 0; i < messages.length; i++) {
		var message = messages[i];
		var displayText = this.formatDisplayMessage(message);
		var messageData = {};
		messageData.id = dataKey;
		messageData.commandNum = i;
		Window_Command.prototype.addCommand.call(this, displayText, messageData);
	}
};

// Make list of messages in troop, with unselectable headers for page.
TextDebugger_Window_ChoiceList.prototype.makeTroopMessageList = function(event) {
	var language = ConfigManager.getLanguage();
	CTE.loadLanguage(language);

	var file = "troops";
	if (typeof CTE.data[language][file] === "undefined") return;

	var keys = Object.keys(CTE.data[language][file]);
	var lastHeader = "";
	for (var i = 0; i < keys.length; i++) {
		var dataKey = keys[i];
		if (!dataKey.contains("-message")) continue;

		var firstDashIndex = dataKey.indexOf("-");
		var secondDashIndex = dataKey.indexOf("-", firstDashIndex + 1);
		var eventID = dataKey.substring(0, firstDashIndex);
		var pageID = dataKey.substring(firstDashIndex + 1, secondDashIndex);
		if (secondDashIndex != -1 && eventID == event) {
			var header = "\\c[1]***** Page " + pageID + " *****";
			if (header !== lastHeader) {
				Window_Command.prototype.addCommand.call(this, header, "");
				lastHeader = header;
			}

			var messages = CTE.getDataForLanguage(language, file, dataKey);
			for (var k = 0; k < messages.length; k++) {
				var message = messages[k];
				var displayText = this.formatDisplayMessage(message);
				var messageData = {};
				messageData.id = dataKey;
				messageData.commandNum = k;
				Window_Command.prototype.addCommand.call(this, displayText, messageData);
			}
		}
	}
};

// Modify message for display in choice list as a single line.
TextDebugger_Window_ChoiceList.prototype.formatDisplayMessage = function(message) {
	var tempMessage = CTE.messageDebuggerPreviewTextHandler(message);
	if (tempMessage) message = tempMessage;

	if (Array.isArray(message)) { // Conditional text array; go with text from first conditional branch
		message = message[0][1];
	}

	message = message.replace(/\n/g, " "); // Replace linebreaks
	message = message.replace(/\\}/g, ""); // Remove \} (shrink text) codes to avoid text being parsed as "{...}" and showing as undefined
	message = this.convertEscapeCharacters(message); // Process remaining escape codes
	return message.trim();
};

// Return size of list.
TextDebugger_Window_ChoiceList.prototype.maxItems = function() {
	return this._list.length;
};

// Draw choice entry.
TextDebugger_Window_ChoiceList.prototype.drawItem = function(index) {
	const rect = this.itemRect(index);
	const text = this._list[index].name;

	// Calculate maximum font size of text to determine Y offset.
	this.contents.fontSize = 20;
	var maxFontSize = 20;
	var regExp = /\x1b[\{\}]/g;
	for (; ;) {
		var array = regExp.exec(text);
		if (array) {
			if (array[0] === "\x1b{") {
				this.makeFontBigger();
			}
			if (array[0] === "\x1b}") {
				this.makeFontSmaller();
			}
			if (maxFontSize < this.contents.fontSize) {
				maxFontSize = this.contents.fontSize;
			}
		}
		else {
			break;
		}
	}
	this.contents.fontSize = 20;

	var heightOffset = (20 - maxFontSize) / 2;
	this.drawTextEx(text, rect.x, rect.y + heightOffset);
};

// Process input.
TextDebugger_Window_ChoiceList.prototype.update = function() {
	this._animationCount++;

	Window_Selectable.prototype.processCursorMove.call(this);

	// Allow holding down left/right/page buttons to scroll.
	var lastIndex = this.index();
	if (Input.isRepeated("right")) {
		this.select(Math.min(this.index() + 10, this.maxItems() - 1));
	}
	if (Input.isRepeated("left")) {
		this.select(Math.max(this.index() - 10, 0));
	}
	if (this.index() !== lastIndex) {
		SoundManager.playCursor();
	}

	if (Input.isRepeated("pagedown")) {
		this.cursorPagedown();
	}
	if (Input.isRepeated("pageup")) {
		this.cursorPageup();
	}

	if (Input.isTriggered("messagedebug")) {
		var searchKey = window.prompt("Enter text to search for.", CTE.TextDebug.searchString);
		if (searchKey === null || searchKey === "") return;
		CTE.TextDebug.searchString = searchKey;
		this.doSearch(searchKey);
	}

	if (Input.isTriggered("ok")) {
		if (this.maxItems() == 0) {
			SoundManager.playBuzzer();
			return;
		}
		this.onChoice(this.index());
	}

	if (Input.isTriggered("cancel")) {
		SoundManager.playCancel();
		if (CTE.TextDebug.currentFile !== "") { // Message list; back out and recreate file list
			CTE.TextDebug.lastIndex.set(CTE.TextDebug.getFileIdentifier(), this.index()); // Memorize position in this list
			CTE.TextDebug.currentFile = "";
			CTE.TextDebug.currentEvent = -1;
			this.makeCommandList();
			this.select(CTE.TextDebug.lastFileIndex);
			this.refresh();
		}
		else { // File list; close the menu
			CTE.TextDebug.lastFileIndex = this.index(); // Memorize position in file list
			this.close();
			CTE.TextDebug.open = false;
		}
	}
};

// OK button handler.
TextDebugger_Window_ChoiceList.prototype.onChoice = function(index) {
	var choice = this._list[index].symbol;
	if (choice === "") {
		SoundManager.playBuzzer();
		return;
	}
	SoundManager.playOk();

	if (CTE.TextDebug.currentFile === "") { // File list; enter message list for the selected file
		CTE.TextDebug.lastFileIndex = index; // Memorize position in file list
		if (choice.startsWith("common")) {
			CTE.TextDebug.currentFile = "commonevents";
			CTE.TextDebug.currentEvent = choice.replace("common", "");
		}
		else if (choice.startsWith("troop")) {
			CTE.TextDebug.currentFile = "troops";
			CTE.TextDebug.currentEvent = choice.replace("troop", "");
		}
		else {
			CTE.TextDebug.currentFile = choice;
			CTE.TextDebug.currentEvent = -1;
		}

		var fileIdentifier = CTE.TextDebug.getFileIdentifier();
		if (!CTE.TextDebug.lastIndex.has(fileIdentifier)) CTE.TextDebug.lastIndex.set(fileIdentifier, 0); // Initialize position in this list
		this.makeCommandList();
		this.select(CTE.TextDebug.lastIndex.get(fileIdentifier));
		this.refresh();
	}
	else { // Show selected message, temporarily closing menu (keeping CTE.TextDebug.open set to true)
		CTE.TextDebug.lastIndex.set(CTE.TextDebug.getFileIdentifier(), this.index()); // Memorize position in this list
		this.close();

		CTE.loadLanguage(ConfigManager.getLanguage());
		const source = CTE.TextDebug.currentFile;
		const message = CTE.getData(source, "events", `${choice.eventID}`, "pages", `${(choice.pageID)}`, choice.messageID);
		if (source === "commonevents") source = "commonevent" + CTE.TextDebug.currentEvent;
		if (source === "troops") source = "troop" + CTE.TextDebug.currentEvent;
		CTE.messageDebuggerMessageSettingsHandler(message, source);
		$gameMessage.add(message);
	}
};

// Search through text in list, prioritizing exact matches.
TextDebugger_Window_ChoiceList.prototype.doSearch = function(searchKey) {
	if (this._list.length == 0) return;
	searchKey = searchKey.toLowerCase();

	var startIndex = this.index();
	var index = startIndex;
	var firstMatch = -1;

	do {
		index = (index + 1) % this.maxItems();
		var name = this._list[index].name ? this._list[index].name.toLowerCase() : "";
		if (name === searchKey) {
			this.select(index);
			return;
		}
		else if (firstMatch == -1 && name.includes(searchKey)) {
			firstMatch = index;
		}
	} while (index != startIndex);

	if (firstMatch != -1) {
		this.select(firstMatch);
	}
	else {
		window.alert("Not found.");
	}
};

// Close the menu.
TextDebugger_Window_ChoiceList.prototype.close = function() {
	SceneManager.pop();
};

// After closing a message initiated by message debugger, reopen the menu.
CTE.Window_Message_terminateMessage = Window_Message.prototype.terminateMessage;
Window_Message.prototype.terminateMessage = function() {
	CTE.Window_Message_terminateMessage.apply(this, arguments);
	if (CTE.TextDebug.open || CTE.MessageTyper.previewing) {
		CTE.messageDebuggerMessageSettingsPostHandler();
		if (CTE.TextDebug.open) {
			SceneManager.push(Scene_TextDebugger);
		}
		if (CTE.MessageTyper.previewing) {
			CTE.MessageTyper.previewing = false;
		}
	}
};

/******************************
MESSAGE TYPER
******************************/

CTE.MessageTyper = {};

// Open (or show) the message typer window.
CTE.openMessageTyperWindow = function() {
	if (typeof CTE.MessageTyper.frame !== "undefined") {
		CTE.MessageTyper.frame.focus();
		return;
	}

	if (!Utils.isNwjs()) return;

	var fs = require("fs");
	var path = require("path");
	var gamePath = path.dirname(process.mainModule.filename);
	try {
		fs.writeFileSync(gamePath + "/CTEWindowPlaceholder.htm", "");
	}
	catch (e) {
		CTE.warn("Could not open message typer; unable to make placeholder HTML file.");
		return;
	}

	if (typeof Zale !== "undefined" && typeof Zale.GameFocus !== "undefined") { // If using pause-on-focus plugin, disable for better usability of preview
		CTE.MessageTyper.FocusPauseGraphicsBackup = Zale.GameFocus.PauseGraphics;
		CTE.MessageTyper.FocusPauseAudioBackup = Zale.GameFocus.PauseAudio;
		Zale.GameFocus.PauseGraphics = false;
		Zale.GameFocus.PauseAudio = false;
	}

	var gui = require("nw.gui");
	var width = 800;
	var height = 200;
	var x = Math.round(((window.innerWidth - width) / 2) + window.screenX);
	var y = window.screenY;
	CTE.MessageTyper.gameWindow = gui.Window.get();
	gui.Window.open("CTEWindowPlaceholder.htm", { title: "Message Typer", x: x, y: y, width: width, height: height }, function(newWindow) {
		CTE.MessageTyper.frame = newWindow;
		newWindow.on("loaded", function() { CTE.MessageTyper.initWindow(); });
	});
};

// Initialize the message typer window after creating.
CTE.MessageTyper.initWindow = function() {
	// Remove placeholder file.
	var fs = require("fs");
	fs.unlink("CTEWindowPlaceholder.htm");

	// Unset variable holding window when closed.
	CTE.MessageTyper.frame.on("close", function() {
		CTE.MessageTyper.frame = undefined;
		if (typeof CTE.MessageTyper.FocusPauseGraphicsBackup !== "undefined") {
			Zale.GameFocus.PauseGraphics = CTE.MessageTyper.FocusPauseGraphicsBackup;
			Zale.GameFocus.PauseAudio = CTE.MessageTyper.FocusPauseAudioBackup;
		}
		this.close(true);
	});

	// Close window when game window closes/is reloaded.
	CTE.MessageTyper.gameWindow.on("document-end", function() {
		if (typeof CTE.MessageTyper.frame !== "undefined") CTE.MessageTyper.frame.close();
	});
	CTE.MessageTyper.gameWindow.on("close", function() {
		if (typeof CTE.MessageTyper.frame !== "undefined") CTE.MessageTyper.frame.close();
		this.close(true);
	});

	// Set window to always be on top.
	CTE.MessageTyper.frame.setAlwaysOnTop(true);

	// Set up document body.
	var document = CTE.MessageTyper.frame.window.document;
	document.body.style.height = (CTE.MessageTyper.frame.height - 16) + "px";

	// Load font CSS.
	var cssLink = document.createElement("link");
	cssLink.rel = "stylesheet";
	cssLink.type = "text/css";
	cssLink.href = "fonts/gamefont.css";
	document.getElementsByTagName("head")[0].appendChild(cssLink);

	// Create input box.
	var input = document.createElement("textarea");
	input.style.width = "100%";
	input.style.height = (CTE.MessageTyper.frame.height - 48) + "px";
	input.style.fontFamily = "GameFont";
	input.style.fontSize = "24px";
	input.style.resize = "none";
	input.onkeypress = CTE.MessageTyper.onKeyPress;
	input.oninput = CTE.MessageTyper.onInput;
	document.body.appendChild(input);
	input.focus();
	CTE.MessageTyper.inputBox = input;

	// Create option boxes.
	CTE.MessageTyper.speedOption = [];
	for (var i = 0; i < 2; i++) {
		var displaySpeed = document.createElement("input");
		displaySpeed.type = "radio";
		displaySpeed.name = "displaySpeed";
		displaySpeed.id = i == 0 ? "instant" : "normal";
		displaySpeed.value = i == 0 ? "instant" : "normal";
		if (i == 0) displaySpeed.checked = true;
		displaySpeed.style.float = "left";
		document.body.appendChild(displaySpeed);
		CTE.MessageTyper.speedOption[i] = displaySpeed;

		var speedLabel = document.createElement("label");
		speedLabel.style.float = "left";
		speedLabel.style.marginTop = "10px";
		speedLabel.style.fontFamily = "sans-serif";
		speedLabel.style.userSelect = "none";
		speedLabel.innerHTML = i == 0 ? "Show Instantly" : "Type Normally";
		speedLabel.appendChild(displaySpeed);
		document.body.appendChild(speedLabel);
	}

	var padding = document.createElement("div");
	padding.style.display = "inline-block";
	padding.style.padding = "32px 32px 0px 0px";
	padding.style.float = "left";
	document.body.appendChild(padding);

	CTE.MessageTyper.updateOption = [];
	for (var i = 0; i < 2; i++) {
		var updateType = document.createElement("input");
		updateType.type = "radio";
		updateType.name = "updateType";
		updateType.id = i == 0 ? "instant" : "normal";
		updateType.value = i == 0 ? "instant" : "normal";
		if (i == 0) updateType.checked = true;
		updateType.style.float = "left";
		document.body.appendChild(updateType);
		CTE.MessageTyper.updateOption[i] = updateType;

		var updateLabel = document.createElement("label");
		updateLabel.style.float = "left";
		updateLabel.style.marginTop = "10px";
		updateLabel.style.fontFamily = "sans-serif";
		updateLabel.style.userSelect = "none";
		updateLabel.innerHTML = i == 0 ? "Refresh On Edit" : "Don't Refresh";
		updateLabel.appendChild(updateType);
		document.body.appendChild(updateLabel);
	}

	// Create preview button.
	var previewButton = document.createElement("button");
	previewButton.innerHTML = "Preview";
	previewButton.style.fontSize = "16px";
	previewButton.style.float = "right";
	previewButton.style.marginTop = "8px";
	previewButton.style.userSelect = "none";
	previewButton.onclick = CTE.MessageTyper.startPreview;
	document.body.appendChild(previewButton);

	// Create message terminate button.
	var terminateButton = document.createElement("button");
	terminateButton.innerHTML = "Close Preview";
	terminateButton.style.fontSize = "16px";
	terminateButton.style.float = "right";
	terminateButton.style.marginTop = "8px";
	terminateButton.style.userSelect = "none";
	terminateButton.onclick = CTE.MessageTyper.terminatePreview;
	document.body.appendChild(terminateButton);
};

// On pressing Enter, preview the typed message in-game.
CTE.MessageTyper.onKeyPress = function(event) {
	if (event.keyCode == 13 && !event.shiftKey) { // Enter
		CTE.MessageTyper.startPreview();
		event.preventDefault();
	}
};

// Close message window if previewing.
CTE.MessageTyper.terminatePreview = function() {
	if (CTE.MessageTyper.previewing && $gameMap && $gameMessage) {
		SceneManager._scene._messageWindow.terminateMessage(); // Close message window
		SceneManager._scene._messageWindow.pause = false; // Clear waiting for input
		SceneManager._scene._messageWindow._waitCount = 0; // Clear pauses
		$gameMessage.clear();
	}
};

// If textbox changes while previewing a message and setting is enabled, reset message with new text.
CTE.MessageTyper.onInput = function(event) {
	if (CTE.MessageTyper.updateOption[0].checked && CTE.MessageTyper.previewing && $gameMap && $gameMessage) {
		CTE.MessageTyper.terminatePreview();

		var message = CTE.MessageTyper.inputBox.value;
		CTE.messageDebuggerMessageSettingsHandler(message, "");
		$gameMessage.add(message);
		SceneManager._scene._messageWindow.startMessage();
		SceneManager._scene._messageWindow._showFast = CTE.MessageTyper.speedOption[0].checked;
		CTE.MessageTyper.previewing = true; // Reset after it gets turned off by terminating message
	}
};

// Preview the typed message in-game.
CTE.MessageTyper.startPreview = function() {
	if ($gameMessage && !$gameMessage.isBusy()) {
		CTE.MessageTyper.previewing = true;
		var message = CTE.MessageTyper.inputBox.value;
		CTE.messageDebuggerMessageSettingsHandler(message, "");
		$gameMessage.add(message);
		SceneManager._scene._messageWindow.startMessage();
		SceneManager._scene._messageWindow._showFast = CTE.MessageTyper.speedOption[0].checked;
	}
};

/******************************
WARNING LOGS
******************************/

CTE.warn = function(object, level = 1) {
	if (CTE.LogSetting >= level) {
		console.log(object);
	}
};

/******************************
CUSTOMIZABLE HANDLERS
******************************/

// Customizable directory for language files. Can be defined before the plugin loads.
if (typeof CTE_getLanguagesFolder === "undefined") {
	CTE_getLanguagesFolder = function() {
		return "languages/";
	};
}

// Customizable handler for command export. Check command.code for any desired commands, and return a string to write.
// If dealing with a command code that already writes something, this "prefix" function lets you insert text before the standard output.
CTE.commandExportHandlerPrefix = function(command) {
	if (command.code == 99999) { // Placeholder
		return "{Comment}\n\n";
	}
	return ""; // Nothing to write
};

// If dealing with a command code that already writes something, this "postfix" function lets you insert text after the standard output.
// (Unless you want to write something both before and after an already-supported command, you probably only want to use one or the other.)
CTE.commandExportHandlerPostfix = function(command) {
	if (command.code == 99999) { // Placeholder
		return "{Comment}\n\n";
	}
	return ""; // Nothing to write
};

// Called whenever the language is changed or language files are reloaded.
// Use it to refresh anything that uses translated text and may be open when language is changed.
CTE.languageRefreshHandler = function() {
};

// Called at the start of the function that processes display text for messages in the Message Debugger.
// Useful for replacing message codes added by plugins.
CTE.messageDebuggerPreviewTextHandler = function(message) {
	//message = message.replace(/\\yourcodehere\[[^\]+]\]/gi, "");
	return message;
};

// Called before displaying messages via the Message Debugger. Use to configure message display (i.e. window visibility) as desired.
// Passes message content and name of source ("map##", "commonevent##", or "troop##") for potential reference purposes.
CTE.messageDebuggerMessageSettingsHandler = function(message, source) {
	$gameMessage.setBackground(0); // 0 for window, 1 for dimmed box, 2 for transparent
};

// Called on returning from message displayed via the Message Debugger. Useful for erasing any message window pictures displayed beforehand.
CTE.messageDebuggerMessageSettingsPostHandler = function() {
};

// Called for default language, prioritized over Default Language setting.
// The default function tries to returns a language code based on the Steam user language if using a Steam API plugin.
// Change the language codes here if they don't match the ones you're using. Theoretically modifiable to use the API of another launcher.
CTE.languageCodeFromExternalAPI = function() {
	// Set languageFunction to a known function that will return Steam user language; if none is found, set to null.
	var languageFunction = (typeof OrangeGreenworks !== "undefined" ? OrangeGreenworks.getGameLanguage : undefined) // OrangeGreenworks
		|| Game_System.prototype.getCurrentGameLanguage // Archeia_Steamworks
		|| null;

	// No function found, so return empty string.
	if (languageFunction === null) {
		return "";
	}

	// Get Steam API language code using the chosen function.
	var userLanguage = languageFunction();
	if (typeof userLanguage === "undefined" || userLanguage === null) {
		return "";
	}

	// Convert userLanguage (Steam API language code) to languageCode (language code for this plugin).
	var languageCode = "";
	switch (userLanguage) {
		case "arabic": languageCode = "ar"; break;
		case "bulgarian": languageCode = "bg"; break;
		case "schinese": languageCode = "zh-CN"; break;
		case "tchinese": languageCode = "zh-TW"; break;
		case "czech": languageCode = "cs"; break;
		case "danish": languageCode = "da"; break;
		case "dutch": languageCode = "nl"; break;
		case "english": languageCode = "en"; break;
		case "finnish": languageCode = "fi"; break;
		case "french": languageCode = "fr"; break;
		case "german": languageCode = "de"; break;
		case "greek": languageCode = "el"; break;
		case "hungarian": languageCode = "hu"; break;
		case "italian": languageCode = "it"; break;
		case "japanese": languageCode = "jp"; break;
		case "koreana": languageCode = "ko"; break;
		case "norwegian": languageCode = "no"; break;
		case "polish": languageCode = "pl"; break;
		case "portuguese": languageCode = "pt"; break;
		case "brazilian": languageCode = "pt-BR"; break;
		case "romanian": languageCode = "ro"; break;
		case "russian": languageCode = "ru"; break;
		case "spanish": languageCode = "es"; break;
		case "latam": languageCode = "es-419"; break;
		case "swedish": languageCode = "sv"; break;
		case "thai": languageCode = "th"; break;
		case "turkish": languageCode = "tr"; break;
		case "ukrainian": languageCode = "uk"; break;
		case "vietnamese": languageCode = "vn"; break;
	}

	// If a languageCode was determined, check that the language code is in the Language Codes list before returning it.
	if (languageCode !== "") {
		languageCode = languageCode.toUpperCase();
		return LanguageCodes.find(code => languageCode === code.toUpperCase()) || "";
	}

	// No valid code found, so return empty string.
	return "";
};